import numpy as np
import pandas as pd

from scipy.interpolate import interp1d

import sqlite3
import matplotlib.pyplot as plt
try:
    from Deformation_Method import DeformationMethod
    from Cross_Section_Properties import cross_section_polygon
except:
    from KIB_LAP.Dynamik.Deformation_Method import DeformationMethod
    from KIB_LAP.Dynamik.Cross_Section_Properties import cross_section_polygon  


class Balkenschwingungen:
    def __init__(
        self,
        _l=16.40,
        _A=1.00,
        _rho=5063,
        _xi=0.0257,
        _E=2.1e11,
        _I=0.0288,
        _t_ber = 5,
        _BC="Hinged-Hinged",
        _cross_section=None,
        _x0 = 16.4/2,
        _v_train = 300
    ):
        ## Calculation Parameters

        self.ndt = 20  # Anzahl der Unterteilungen der kleinsten

        self.x0 =  _x0  # Stelle der Ausgabe [m]

        self.n = 5  # Anzahl der berücksichtigten Eigenformen

        self.v_zug_kmh = _v_train # Zuggeschwindigkeit in [km/h]

        ## Parameters for the pedestrian force vectors

        self.v_pedestrian = 2.205  # Speed of the pedestrian walk in [m/s]

        # Calculation time for harmonic and pedestrian calculation

        self.t_ber = _t_ber  # Dauer der Berechnung
        self.dt = 1e-3  # Definition of the time step

        # Integrationsparameter
        self.alpha = 0.5
        self.beta = 0.25

        # Voreinstellung der Lagerbedingungen

        self.boundary_condition = _BC

        # System properties

        self.l = _l  # Trägerlänge [m]
        if _cross_section == None:
            self.A = _A
            self.I = _I  # Flächenträgheitsmoment [m^4]
        else:
            self.A = _cross_section.A
            self.I = _cross_section.I_yy
        self.rho = _rho
        self.mue = self.A * self.rho  # Masse pro Längeneinheit [kg/m]
        self.xi = _xi  # Dämpfungsmaß [-]

        self.E = _E  # Elastizitätsmodul [N/m^2]

        # Loading condition

        self.loading = "train"

        #   Harmonic-Loading

        self.OMEGA = 6.38 * 2 * np.pi
        self.PHI_0 = 0
        self.F_0_H = 1e5

        # Test function

        self.test = False
        # Functions

        if self.test == True:
            self.load_database()
            self.load_definition()
            self.system_properties()
            self.load_conditions()
            self.generalized_load_function()
            self.beta_newmark()

        else:
            self.load_database()
            self.load_definition()
            self.system_properties()
            self.load_conditions()
            self.generalized_load_function()
            self.beta_newmark()

    def load_database(self):
        db_name = "Database/eigenmodes.db"
        # Connect to the SQLite database
        self.conn = sqlite3.connect(db_name)
        self.cursor = self.conn.cursor()

        # Initializing numpy arrays
        self.lambda_cantilever = np.zeros(20)
        self.lambda_clamped_hinged = np.zeros(20)
        self.lambda_hinged_hinged = np.zeros(20)
        self.lambda_clamped_clamped = np.zeros(20)

        self.lambda_hinged_hinged_twospan = np.zeros(20)

        self.a_lambda_cantilever = np.zeros(20)
        self.a_lambda_clamped_hinged = np.zeros(20)
        self.a_lambda_clamped_clamped = np.zeros(20)

        self.J1_cantilever = np.zeros(20)
        self.J1_clamped_hinged = np.zeros(20)
        self.J1_hinged_hinged = np.zeros(20)
        self.J1_clamped_clamped = np.zeros(20)

        self.J2_cantilever = np.zeros(20)
        self.J2_clamped_hinged = np.zeros(20)
        self.J2_hinged_hinged = np.zeros(20)
        self.J2_clamped_clamped = np.zeros(20)

        self.J3_cantilever = np.zeros(20)
        self.J3_clamped_hinged = np.zeros(20)
        self.J3_hinged_hinged = np.zeros(20)
        self.J3_clamped_clamped = np.zeros(20)

        # Saving the Data-Base into the numpy arrays

        self.cursor.execute(
            "SELECT lambda_cantilever, lambda_clamped_hinged, lambda_hinged_hinged, lambda_clamped_clamped FROM eigenvalues"
        )
        data = self.cursor.fetchall()
        for i, row in enumerate(data):
            (
                self.lambda_cantilever[i],
                self.lambda_clamped_hinged[i],
                self.lambda_hinged_hinged[i],
                self.lambda_clamped_clamped[i],
            ) = row

        self.cursor.execute(
            "SELECT a_lambda_cantilever, a_lambda_clamped_hinged, a_lambda_clamped_clamped FROM a_lambda"
        )
        data = self.cursor.fetchall()
        for i, row in enumerate(data):
            (
                self.a_lambda_cantilever[i],
                self.a_lambda_clamped_hinged[i],
                self.a_lambda_clamped_clamped[i],
            ) = row

        self.cursor.execute(
            "SELECT J1_cantilever, J1_clamped_hinged, J1_hinged_hinged, J1_clamped_clamped FROM J1"
        )
        data = self.cursor.fetchall()
        for i, row in enumerate(data):
            (
                self.J1_cantilever[i],
                self.J1_clamped_hinged[i],
                self.J1_hinged_hinged[i],
                self.J1_clamped_clamped[i],
            ) = row

        self.cursor.execute(
            "SELECT J2_cantilever, J2_clamped_hinged, J2_hinged_hinged, J2_clamped_clamped FROM J2"
        )
        data = self.cursor.fetchall()
        for i, row in enumerate(data):
            (
                self.J2_cantilever[i],
                self.J2_clamped_hinged[i],
                self.J2_hinged_hinged[i],
                self.J2_clamped_clamped[i],
            ) = row

        self.cursor.execute(
            "SELECT J3_cantilever, J3_clamped_hinged, J3_hinged_hinged, J3_clamped_clamped FROM J3"
        )
        data = self.cursor.fetchall()
        for i, row in enumerate(data):
            (
                self.J3_cantilever[i],
                self.J3_clamped_hinged[i],
                self.J3_hinged_hinged[i],
                self.J3_clamped_clamped[i],
            ) = row

        for i in range(1,len(data)+1):
            if (i%2==1):
                self.lambda_hinged_hinged_twospan[i-1] = self.lambda_hinged_hinged[(i-1)//2]
            else:
                self.lambda_hinged_hinged_twospan[i-1] = self.lambda_clamped_hinged[(i//2)-1]

        print("Lambda - Two Span")
        print(self.lambda_hinged_hinged_twospan)
        # Close the connection
        self.conn.close()

    def load_definition(self):
        if self.loading == "train":
            Kraftdefinition = pd.read_csv(
                "Trainpassing/Inputdatei_1.txt", delim_whitespace=True
            )
            # # # Coordinates related to the first pair of wheels in [m]
            self.x_k = Kraftdefinition.iloc[:, 0].to_list()
            # # # Axial loads in [kN]
            self.P_k = Kraftdefinition.iloc[:, 1].to_list()

            self.P_k_array = np.zeros(len(self.P_k))

            for i in range(0, len(self.P_k), 1):
                self.P_k_array[i] = self.P_k[i]

        elif self.loading == "harmonic_single":
            self.x_k = self.l / 2
            self.P_k = 1

    def numeric_eigenvalues(self):
        # Calculation for numeric eigenmodes
        if self.boundary_condition == "Hinged-Hinged-Numeric":
            self.Deformation_Num = DeformationMethod(
                30, self.E, self.I, self.A, self.rho, self.l, "SPG", [1, 1], 1000
            )
            self.Deformation_Num.single_span_girder()
            self.Deformation_Num.compute_eigenfrequencies()
            self.Deformation_Num.modal_matrices()
        elif self.boundary_condition == "Two-Span-Girder":
            self.Deformation_Num = DeformationMethod(
                30, self.E, self.I, self.A, self.rho, self.l, "SPG", [1, 1], 1000
            )
            self.Deformation_Num.single_span_girder()
            self.Deformation_Num.compute_eigenfrequencies()
            self.Deformation_Num.modal_matrices()

    def load_conditions(self):
        # Eigenformen am Ausgabepunkt
        self.phi_a = np.zeros(self.n)

        # Loop for defining the eigenmodes at the loading point
        for i in range(self.n):
            if self.boundary_condition == "Hinged-Hinged":
                self.phi_a[i] = self.hinged_hinged(self.x0, i)
            elif self.boundary_condition == "Hinged-Hinged-Numeric":
                self.phi_a[i] = self.hinged_hinged_numeric(self.x0, i)
            elif self.boundary_condition == "Clamped-Hinged":
                self.phi_a[i] = self.clamped_hinged(self.x0, i)
            elif self.boundary_condition == "Clamped-Clamped":
                self.phi_a[i] = self.clamped_clamped(self.x0, i)
            elif self.boundary_condition == "Cantilever":
                self.phi_a[i] = self.cantilever(self.x0, i)
            elif self.boundary_condition == "Hinged-Hinged-TwoSpan":
                self.phi_a[i] = self.hingedhinged_twospan(self.x0, i)

        if self.loading == "train":
            # Train-Passing
            self.v_zug = self.v_zug_kmh * 1000 / 3600  # Zuggeschwindigkeit in [m/s]
            self.dt = 1e-3
            self.l_zug = max(self.x_k)  # Zuglänge
            self.T_u = (self.l + self.l_zug) / (
                self.v_zug
            )  # Zeit zur Überquerung des Trägers

            self.nt = int(np.ceil(self.t_ber / self.dt) + 1)  # Anzahl der Zeitschritte

            # Erstellung des Zeitvektors
            self.t = np.arange(0, self.nt * self.dt, self.dt)

            # Kraftdefinition für die Berechnung

            # Matrix zur Berücksichtigung der sich auf der Brücke befindenden Lasten
            # in jedem Zeitschritt

            self.F_Mat = np.zeros(
                (len(self.x_k), self.nt)
            )  # rows -> Index of the train load, cols = number of time steps

            print(max(self.t))

            for i in range(self.nt):
                for j in range(len(self.x_k)):
                    if (-self.x_k[j] + self.v_zug * self.t[i] > 0) and (
                        -self.x_k[j] + self.v_zug * self.t[i] < self.l
                    ):  # Condition, that the train needs to be on the bridge
                        self.F_Mat[j][i] = self.P_k[j]
                    else:
                        self.F_Mat[j][i] = 0

        elif self.loading == "harmonic_single":
            self.nt = int(np.ceil(self.t_ber / self.dt) + 1)  # Anzahl der Zeitschritte
            self.t = np.arange(0, self.nt * self.dt, self.dt)

            self.F_Mat = np.zeros(
                self.nt
            )  # rows = number of time steps, because there is just one single load

            for i in range(self.nt):
                self.F_Mat[i] = self.F_0_H * np.sin(
                    self.OMEGA * self.dt * i - self.PHI_0
                )

        elif self.loading == "pedestrian":
            self.nt = int(np.ceil(self.t_ber / self.dt) + 1)  # Anzahl der Zeitschritte
            self.t = np.arange(0, self.nt * self.dt, self.dt)

            self.F_Mat = np.zeros(
                self.nt
            )  # rows = number of time steps, because there is just one single load

            g_pedestrian = 700  # 800 N for normal case
            coeff_1 = 0.50
            coeff_2 = 0.10
            coeff_3 = 0.10

            step = 0.90

            for i in range(self.nt):
                if (self.v_pedestrian * self.t[i] > 0) and (
                    self.v_pedestrian * self.t[i] < self.l
                ):
                    self.F_Mat[i] = (
                        1
                        + coeff_1 * np.sin(2 * np.pi * 2.45 * self.t[i])
                        + coeff_2 * np.sin(4 * np.pi * 2.45 * self.t[i] - np.pi / 2)
                        + coeff_3 * np.sin(6 * np.pi * 2.45 * self.t[i] - np.pi / 2)
                    ) * g_pedestrian
                else:
                    self.F_Mat[i] = 0

    def system_properties(self):
        # Berechnung weiterer System- und Berechnungsparameter
        self.Freq = np.zeros(self.n)  # Eigenfrequenzen
        self.Omega = np.zeros(self.n)  # Eigenkreisfrequenzen
        self.T = np.zeros(self.n)  # Eigenschwingzeiten

        if self.boundary_condition == "Hinged-Hinged":
            for i in range(0, self.n, 1):
                if i == 0:
                    self.Freq[i] = i  # Static movement due to the
                else:
                    self.Freq[i] = (
                        (self.lambda_hinged_hinged[i - 1]) ** 2
                        / (2 * np.pi * self.l**2)
                    ) * np.sqrt(self.E * self.I / self.mue)
                print(self.Freq[i])

                self.Omega[i] = 2 * np.pi * self.Freq[i]
                if i == 0:
                    self.T[i] = 1e9
                else:
                    self.T[i] = 1 / self.Freq[i]

            self.m = np.zeros(self.n)  # Vektor der modalen Massen
            self.k = np.zeros(self.n)  # Vektor der modalen Steifigkeiten
            self.d = np.zeros(self.n)  # Vektor der modalen Dämpferkonstanten

            for i in range(self.n):
                self.m[i] = self.J1_hinged_hinged[i] * self.mue * self.l
                self.k[i] = self.Omega[i] ** 2 * self.m[i]
                self.d[i] = 2 * self.xi * np.sqrt(self.k[i] * self.m[i])

                print(self.m)

        elif self.boundary_condition == "Hinged-Hinged-Numeric":
            print("System Properties")

            for i in range(0, self.n, 1):
                self.Freq[i] = self.Deformation_Num.eigenfrequencies[i] / (np.pi * 2)
                self.Omega[i] = self.Deformation_Num.eigenfrequencies[i]
                self.T[i] = 1 / self.Freq[i]

            print(self.Deformation_Num.M_trans)
            print(self.Deformation_Num.K_trans)

            self.m = np.zeros(self.n)  # Vektor der modalen Massen
            self.k = np.zeros(self.n)  # Vektor der modalen Steifigkeiten
            self.d = np.zeros(self.n)  # Vektor der modalen Dämpferkonstanten

            for i in range(self.n):
                self.m[i] = self.Deformation_Num.M_trans[i][i]
                self.k[i] = self.Deformation_Num.K_trans[i][i]
                self.d[i] = 2 * self.xi * np.sqrt(self.k[i] * self.m[i])

        elif self.boundary_condition == "Clamped-Hinged":
            for i in range(0, self.n, 1):
                if i == 0:
                    self.Freq[i] = i  # Static movement ? Check, if necessary
                else:
                    self.Freq[i] = (
                        (self.lambda_clamped_hinged[i - 1]) ** 2
                        / (2 * np.pi * self.l**2)
                    ) * np.sqrt(self.E * self.I / self.mue)
                print(self.Freq[i])

                self.Omega[i] = 2 * np.pi * self.Freq[i]
                if i == 0:
                    self.T[i] = 1e9
                else:
                    self.T[i] = 1 / self.Freq[i]

            self.m = np.zeros(self.n)  # Vektor der modalen Massen
            self.k = np.zeros(self.n)  # Vektor der modalen Steifigkeiten
            self.d = np.zeros(self.n)  # Vektor der modalen Dämpferkonstanten

            for i in range(self.n):
                self.m[i] = self.J1_clamped_hinged[i] * self.mue * self.l
                self.k[i] = self.Omega[i] ** 2 * self.m[i]
                self.d[i] = 2 * self.xi * np.sqrt(self.k[i] * self.m[i])

                print(self.m)

        elif self.boundary_condition == "Clamped-Clamped":
            for i in range(0, self.n, 1):
                if i == 0:
                    self.Freq[i] = i  # Static movement due to the
                else:
                    self.Freq[i] = (
                        (self.lambda_clamped_clamped[i - 1]) ** 2
                        / (2 * np.pi * self.l**2)
                    ) * np.sqrt(self.E * self.I / self.mue)
                print(self.Freq[i])

                self.Omega[i] = 2 * np.pi * self.Freq[i]
                if i == 0:
                    self.T[i] = 1e9
                else:
                    self.T[i] = 1 / self.Freq[i]

            self.m = np.zeros(self.n)  # Vektor der modalen Massen
            self.k = np.zeros(self.n)  # Vektor der modalen Steifigkeiten
            self.d = np.zeros(self.n)  # Vektor der modalen Dämpferkonstanten

            for i in range(self.n):
                self.m[i] = self.J1_clamped_clamped[i] * self.mue * self.l
                self.k[i] = self.Omega[i] ** 2 * self.m[i]
                self.d[i] = 2 * self.xi * np.sqrt(self.k[i] * self.m[i])

                print(self.m)

        elif self.boundary_condition == "Cantilever":
            for i in range(0, self.n, 1):
                if i == 0:
                    self.Freq[i] = i  # Static movement due to the
                else:
                    self.Freq[i] = (
                        (self.lambda_cantilever[i - 1]) ** 2 / (2 * np.pi * self.l**2)
                    ) * np.sqrt(self.E * self.I / self.mue)
                print(self.Freq[i])

                self.Omega[i] = 2 * np.pi * self.Freq[i]
                if i == 0:
                    self.T[i] = 1e9
                else:
                    self.T[i] = 1 / self.Freq[i]

            self.m = np.zeros(self.n)  # Vektor der modalen Massen
            self.k = np.zeros(self.n)  # Vektor der modalen Steifigkeiten
            self.d = np.zeros(self.n)  # Vektor der modalen Dämpferkonstanten

            for i in range(self.n):
                self.m[i] = self.J1_cantilever[i] * self.mue * self.l
                self.k[i] = self.Omega[i] ** 2 * self.m[i]
                self.d[i] = 2 * self.xi * np.sqrt(self.k[i] * self.m[i])

                print(self.m)

    def hinged_hinged(self, x_load, n_eigen):
        if n_eigen == 0:
            return np.sin(x_load * 0 / self.l)
        else:
            return np.sin(x_load * self.lambda_hinged_hinged[n_eigen - 1] / self.l)

    def hinged_hinged_numeric(self, x_load, n_eigen):
        interpolator = interp1d(
            self.Deformation_Num.len_plotting,
            self.Deformation_Num.eigenmodes_matrix[:, n_eigen],
            kind="linear",
            fill_value=0,
        )
        interpolated_value = interpolator(x_load)
        return interpolated_value

    def clamped_hinged(self, x_load, n_eigen):
        if n_eigen == 0:
            value = 0
        else:
            value = (
                np.sin(self.lambda_clamped_hinged[n_eigen - 1] * x_load / self.l)
                - np.sinh(self.lambda_clamped_hinged[n_eigen - 1] * x_load / self.l)
                + self.a_lambda_clamped_hinged[n_eigen - 1]
                * (
                    np.cosh(self.lambda_clamped_hinged[n_eigen - 1] * x_load / self.l)
                    - np.cos(self.lambda_clamped_hinged[n_eigen - 1] * x_load / self.l)
                )
            )
        return value

    def clamped_clamped(self, x_load, n_eigen):
        if n_eigen == 0:
            value = 0
        else:
            value = (
                np.sin(self.lambda_clamped_clamped[n_eigen - 1] * x_load / self.l)
                - np.sinh(self.lambda_clamped_clamped[n_eigen - 1] * x_load / self.l)
                + self.a_lambda_clamped_clamped[n_eigen - 1]
                * (
                    np.cosh(self.lambda_clamped_clamped[n_eigen - 1] * x_load / self.l)
                    - np.cos(self.lambda_clamped_clamped[n_eigen - 1] * x_load / self.l)
                )
            )
        return value

    def cantilever(self, x_load, n_eigen):
        if n_eigen == 0:
            value = 0
        else:
            value = (
                np.sin(self.lambda_cantilever[n_eigen - 1] * x_load / self.l)
                - np.sinh(self.lambda_cantilever[n_eigen - 1] * x_load / self.l)
                + self.a_lambda_cantilever[n_eigen - 1]
                * (
                    np.cosh(self.lambda_cantilever[n_eigen - 1] * x_load / self.l)
                    - np.cos(self.lambda_cantilever[n_eigen - 1] * x_load / self.l)
                )
            )
        return value

    def hingedhinged_twospan(self,x_load,n_eigen):
        if n_eigen == 0:
            return 0
        else:
            if (n_eigen %2 == 1):
                lambda_n = self.lambda_hinged_hinged[n_eigen-1]
                return np.sin(x_load * lambda_n / self.l)
            else:
                lambda_n = self.lambda_clamped_hinged[n_eigen-1]
                return np.sin(x_load *  lambda_n / self.l)

    def generalized_load_function(self):
        self.Gen_K = np.zeros(
            (self.n, self.nt)
        )  # rows -> Eigenmode 0,...,n , nt = number of the time steps
        if self.loading == "train":
            for u in range(self.n):  # Number of eigenmode
                for i in range(self.nt):    # Number of timesteps
                    for j in range(len(self.x_k)):
                        if self.boundary_condition == "Hinged-Hinged":
                            self.Gen_K[u][i] += self.F_Mat[j][i] * self.hinged_hinged(
                                -self.x_k[j] + self.v_zug * self.t[i], u
                            )
                        elif self.boundary_condition == "Hinged-Hinged-Numeric":
                            try:
                                self.Gen_K[u][i] += self.F_Mat[j][
                                    i
                                ] * self.hinged_hinged_numeric(
                                    -self.x_k[j] + self.v_zug * self.t[i], u
                                )
                            except:
                                self.Gen_K[u][i] += 0
                        elif self.boundary_condition == "Clamped-Hinged":
                            self.Gen_K[u][i] += self.F_Mat[j][i] * self.clamped_hinged(
                                -self.x_k[j] + self.v_zug * self.t[i], u
                            )
                        elif self.boundary_condition == "Clamped-Hinged-Numeric":
                            self.Gen_K[u][i] += self.F_Mat[j][i] * self.clamped_hinged(
                                -self.x_k[j] + self.v_zug * self.t[i], u
                            )
                        elif self.boundary_condition == "Clamped-Clamped":
                            self.Gen_K[u][i] += self.F_Mat[j][i] * self.clamped_clamped(
                                -self.x_k[j] + self.v_zug * self.t[i], u
                            )
                        elif self.boundary_condition == "Cantilever":
                            self.Gen_K[u][i] += self.F_Mat[j][i] * self.cantilever(
                                -self.x_k[j] + self.v_zug * self.t[i], u
                            )

        elif self.loading == "harmonic_single":
            for u in range(self.n):
                for i in range(self.nt):
                    if self.boundary_condition == "Hinged-Hinged":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.hinged_hinged(
                            self.x0, u
                        )
                    elif self.boundary_condition == "Hinged-Hinged-Numeric":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.hinged_hinged_numeric(
                            self.x0, u
                        )
                    elif self.boundary_condition == "Clamped-Hinged":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.clamped_hinged(
                            self.x0, u
                        )
                    elif self.boundary_condition == "Clamped-Clamped":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.clamped_clamped(
                            self.x0, u
                        )
                    elif self.boundary_condition == "Cantilever":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.cantilever(self.x0, u)

        elif self.loading == "pedestrian":
            for u in range(self.n):
                for i in range(self.nt):
                    if self.boundary_condition == "Hinged-Hinged":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.hinged_hinged(
                            self.v_pedestrian * self.t[i], u
                        )
                    elif self.boundary_condition == "Hinged-Hinged-Numeric":
                        try:
                            self.Gen_K[u][i] = self.F_Mat[
                                i
                            ] * self.hinged_hinged_numeric(
                                self.v_pedestrian * self.t[i], u
                            )
                        except:
                            self.Gen_K[u][i] = 0
                    elif self.boundary_condition == "Clamped-Hinged":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.clamped_hinged(
                            self.v_pedestrian * self.t[i], u
                        )
                    elif self.boundary_condition == "Clamped-Clamped":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.clamped_clamped(
                            self.v_pedestrian * self.t[i], u
                        )
                    elif self.boundary_condition == "Cantilever":
                        self.Gen_K[u][i] = self.F_Mat[i] * self.cantilever(
                            self.v_pedestrian * self.t[i], u
                        )

    def beta_newmark(self):
        # Berechnung der Schwingungsantwort mittels Newmark-Verfahren

        # Berechnungsvorschrift
        self.ita_y = np.zeros((self.n, len(self.t)))
        self.ita_v = np.zeros((self.n, len(self.t)))
        self.ita_a = np.zeros((self.n, len(self.t)))

        for j in range(self.n):
            for i in range(1, len(self.t)):
                self.a_h = (
                    ((1 / self.beta) * self.m[j])
                    + (self.alpha / self.beta) * self.d[j] * self.dt
                    + self.k[j] * self.dt**2
                )
                self.b_h = (
                    (
                        (1 / self.beta) * self.m[j]
                        + (self.alpha / self.beta) * self.d[j] * self.dt
                    )
                    * self.ita_y[j][i - 1]
                    + (
                        (1 / self.beta) * self.m[j]
                        + (self.alpha / self.beta - 1) * self.d[j] * self.dt
                    )
                    * self.dt
                    * self.ita_v[j][i - 1]
                    + (
                        (1 / (2 * self.beta) - 1) * self.m[j]
                        + (self.alpha / (2 * self.beta) - 1) * self.d[j] * self.dt
                    )
                    * self.dt**2
                    * self.ita_a[j][i - 1]
                    + self.Gen_K[j][i] * self.dt**2
                )
                self.ita_y[j][i] = self.a_h ** (-1) * self.b_h
                self.ita_v[j][i] = (
                    (self.alpha / (self.beta * self.dt))
                    * (self.ita_y[j][i] - self.ita_y[j][i - 1])
                    - ((self.alpha / self.beta) - 1) * self.ita_v[j][i - 1]
                    - (self.alpha / (2 * self.beta) - 1)
                    * self.dt
                    * self.ita_a[j][i - 1]
                )
                self.ita_a[j][i] = (
                    (1 / (self.beta * self.dt**2))
                    * (self.ita_y[j][i] - self.ita_y[j][i - 1])
                    - 1 / (self.beta * self.dt) * self.ita_v[j][i - 1]
                    - (1 / (2 * self.beta) - 1) * self.ita_a[j][i - 1]
                )

        # Überlagerung aller Eigenformen
        self.y = np.zeros(len(self.t))  # Vektor der Gesamtverschiebung
        self.v = np.zeros(len(self.t))  # Vektor der Gesamtgeschwindigkeit
        self.a = np.zeros(len(self.t))  # Vektor der Gesamtbeschleunigung

        for j in range(self.n):
            for i in range(len(self.t)):
                self.y_tot_h = self.phi_a[j] * self.ita_y[j][i]
                self.y[i] += self.y_tot_h
                self.v_tot_h = self.phi_a[j] * self.ita_v[j][i]
                self.v[i] += self.v_tot_h
                self.a_tot_h = self.phi_a[j] * self.ita_a[j][i]
                self.a[i] += self.a_tot_h

    def extrema(self):
        # Extremwerte
        self.ymax = max(self.y)
        self.ymin = min(self.y)
        self.vmax = max(self.v)
        self.vmin = min(self.v)
        self.amax = max(self.a)
        self.amin = min(self.a)

    def fourier_transformation(self):
        print("Fourier-Transform")

        self.delta_t = self.t[2] - self.t[1]  # Voraussetzung: Äquidistante Zeitschritte
        self.Tges = (
            len(self.t) * self.delta_t
        )  # Gesamtzeit: Letztes Element des Arrays - Erstes Element des Arrays

        self.srate = self.delta_t ** (-1)
        time = self.t
        npnts = len(time)

        # prepare the Fourier transform
        fourTime = np.array(range(npnts)) / npnts
        fCoefs = np.zeros((len(self.y)), dtype=complex)

        for fi in range(npnts):
            # create complex sine wave
            csw = np.exp(-1j * 2 * np.pi * fi * fourTime)
            # compute dot product between sine wave and signal
            # these are called the Fourier coefficients
            fCoefs[fi] = np.sum(np.multiply(self.y, csw)) / npnts

        # Fourier spectrum
        signalX = fCoefs

        self.hz = np.linspace(0, self.srate, npnts)

        print(len(self.hz))

        # amplitude
        self.ampl = np.abs(signalX[0 : len(self.hz)])

        for i in range(0, len(self.hz)):
            if i == 0:
                self.ampl[i] = self.ampl[i]
            else:
                self.ampl[i] = 2 * self.ampl[i]

        plt.stem(self.hz, self.ampl)
        plt.xlim(0, 30)
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Amplitude (a.u.)")
        plt.show()






import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.cm as cm
import matplotlib.colors as colors
from matplotlib.widgets import Slider
from scipy.interpolate import griddata


class ShellPlotter:
    def __init__(self, model):
        """
        model: ShellCalculation Instanz (hat Meshing, AssembleMatrix, stress_elem_avg, ...)
        """
        self.m = model
        self.A = self.m.AssembleMatrix  # Assembled_Matrices

    # ---------- helpers ----------
    def _mesh_bounds(self, pad_rel=0.05, pad_abs=0.0):
        """
        Robuste Auto-Achsen: aus NL min/max + Rand.
        pad_rel: relativer Rand (5% der Größe)
        pad_abs: absoluter Rand (z.B. 0.1 m)
        """
        NL = np.asarray(self.m.Meshing.NL, dtype=float)
        xmin, ymin = NL.min(axis=0)
        xmax, ymax = NL.max(axis=0)
        dx = max(xmax - xmin, 1e-12)
        dy = max(ymax - ymin, 1e-12)
        pad = max(pad_abs, pad_rel * max(dx, dy))
        return xmin - pad, xmax + pad, ymin - pad, ymax + pad

    def _element_coords(self, el):
        return np.array([self.m.Meshing.NL[nid - 1] for nid in el], dtype=float)

    def _point_in_poly(self, x, y, poly):
        """
        Ray casting: True wenn Punkt (x,y) im Polygon (convex/concave) liegt.
        poly: shape (n,2)
        """
        inside = False
        n = len(poly)
        for i in range(n):
            x1, y1 = poly[i]
            x2, y2 = poly[(i + 1) % n]
            cond = ((y1 > y) != (y2 > y)) and (x < (x2 - x1) * (y - y1) / (y2 - y1 + 1e-30) + x1)
            if cond:
                inside = not inside
        return inside

    def _mask_points_in_mesh(self, xy_points):
        """
        xy_points: array shape (N,2)
        returns: bool mask shape (N,) -> True if point lies inside any element polygon
        """
        NL = np.asarray(self.m.Meshing.NL, float)
        EL = np.asarray(self.m.Meshing.EL, int)

        mask = np.zeros((xy_points.shape[0],), dtype=bool)

        # Optional: Precompute polygons once (speed-up)
        polys = []
        for el in EL:
            poly = np.array([NL[nid - 1] for nid in el], dtype=float)
            polys.append(poly)

        for k, (x, y) in enumerate(xy_points):
            inside_any = False
            for poly in polys:
                if self._point_in_poly(float(x), float(y), poly):
                    inside_any = True
                    break
            mask[k] = inside_any

        return mask

    def _mask_cut_points_in_mesh(self, cut_direction, cut_position, coords_1d):
        """
        coords_1d: array der Koordinate entlang der Cut-Achse
                  (y bei x=const, x bei y=const)
        returns: bool mask gleicher Länge, True wenn Punkt in irgendeinem Element liegt
        """
        if cut_direction.lower() == "x":
            xy = np.column_stack([np.full_like(coords_1d, float(cut_position)), coords_1d.astype(float)])
        else:
            xy = np.column_stack([coords_1d.astype(float), np.full_like(coords_1d, float(cut_position))])
        return self._mask_points_in_mesh(xy)

    # ---------- plots ----------
    def plot_mesh(self, show_node_ids=False, show_elem_ids=False, show_springs=True, spring_scale=1.0):
        fig, ax = plt.subplots()
        ax.set_aspect("equal", adjustable="box")
        ax.grid(True)

        # elements
        for e, el in enumerate(self.m.Meshing.EL):
            coords = self._element_coords(el)
            ax.add_patch(patches.Polygon(coords, closed=True, fill=False, edgecolor="r", linewidth=1.0))

            if show_elem_ids:
                c = coords.mean(axis=0)
                ax.text(c[0], c[1], str(e + 1), fontsize=9, ha="center", va="center")

        # nodes
        if show_node_ids:
            NL = np.asarray(self.m.Meshing.NL, float)
            ax.scatter(NL[:, 0], NL[:, 1], s=15)
            for i, (x, y) in enumerate(NL, start=1):
                ax.text(x, y, str(i), fontsize=9, ha="right", va="bottom")

        xmin, xmax, ymin, ymax = self._mesh_bounds()
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        if show_springs:
            self._add_springs_to_ax(ax, spring_scale=spring_scale)

        plt.show()

    def plot_deflected_interactive(self, factor0=1000.0, factor_max=5000.0, show_undeformed=True,
                                  show_springs=True, spring_scale=1.0):
        mesh = self.m.Meshing
        EL = mesh.EL
        NL = mesh.NL

        fig, ax = plt.subplots()
        plt.subplots_adjust(bottom=0.18)
        ax.set_aspect("equal", adjustable="box")
        ax.grid(True)

        xmin, xmax, ymin, ymax = self._mesh_bounds(pad_rel=0.08)
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        # undeformed
        if show_undeformed:
            for el in EL:
                coords = [NL[nid - 1] for nid in el]
                coords.append(coords[0])
                ax.add_patch(patches.Polygon(coords, closed=True, fill=False, edgecolor="0.7", linewidth=1.0))

        # deflected artists
        poly_def = []
        for _ in EL:
            p = patches.Polygon([[0, 0]], closed=True, fill=False, edgecolor="r", linewidth=1.5)
            ax.add_patch(p)
            poly_def.append(p)

        ax_slider = plt.axes([0.15, 0.06, 0.70, 0.03])
        s_factor = Slider(ax_slider, "Scale", 0.0, factor_max, valinit=factor0)

        # prefetch displacements (interleaved per element column)
        Ue = self.m.AssembleMatrix.disp_element_matrix  # shape (8, NoE)

        def _update(factor):
            for e, el in enumerate(EL):
                coords_def = []
                ue = Ue[:, e]
                for local_i, nid in enumerate(el):
                    x0, y0 = NL[nid - 1]
                    ux = ue[2 * local_i + 0]
                    uy = ue[2 * local_i + 1]
                    coords_def.append([x0 + ux * factor, y0 + uy * factor])
                coords_def.append(coords_def[0])
                poly_def[e].set_xy(coords_def)
            fig.canvas.draw_idle()

        s_factor.on_changed(_update)
        _update(factor0)

        if show_springs:
            self._add_springs_to_ax(ax, spring_scale=spring_scale)

        plt.show()

    def plot_inner_element_forces(self, field="sigma_x", show_principal=False, show_springs=True, spring_scale=1.0):
        if not hasattr(self.m, "stress_elem_avg"):
            self.m.CalculateInnerElementForces_Gauss()

        field_idx = {"sigma_x": 0, "sigma_y": 1, "tau_xy": 2,
                     "n_x": 0, "n_y": 1, "n_xy": 2}

        use_n = field.startswith("n_")
        j = field_idx[field]

        vals = (self.m.n_elem_avg[:, j] if use_n else self.m.stress_elem_avg[:, j])
        vals = np.asarray(vals, dtype=float)

        vmin, vmax = float(vals.min()), float(vals.max())
        if np.isclose(vmin, vmax):
            vmin -= 1.0
            vmax += 1.0

        norm = colors.Normalize(vmin=vmin, vmax=vmax)
        cmap = cm.viridis

        fig, ax = plt.subplots()
        ax.set_aspect("equal", adjustable="box")
        ax.grid(True)

        for e, el in enumerate(self.m.Meshing.EL):
            coords = self._element_coords(el)
            polygon = patches.Polygon(coords, closed=True, edgecolor="k", facecolor=cmap(norm(vals[e])))
            ax.add_patch(polygon)

            if show_principal and (not use_n):
                sx, sy, txy = self.m.stress_elem_avg[e, :]
                s_avg = 0.5 * (sx + sy)
                R = np.sqrt((0.5 * (sx - sy)) ** 2 + txy ** 2)
                s1 = s_avg + R
                s2 = s_avg - R
                theta = 0.5 * np.arctan2(2.0 * txy, (sx - sy))

                c = coords.mean(axis=0)
                L = 0.15 * max((coords[:, 0].max() - coords[:, 0].min()),
                               (coords[:, 1].max() - coords[:, 1].min()), 1e-9)
                c1 = "b" if s1 > 0 else "r"
                c2 = "b" if s2 > 0 else "r"
                ax.arrow(c[0], c[1], L * np.cos(theta), L * np.sin(theta),
                         head_width=0.03 * L, head_length=0.03 * L, fc=c1, ec=c1)
                ax.arrow(c[0], c[1], L * np.cos(theta + np.pi / 2), L * np.sin(theta + np.pi / 2),
                         head_width=0.03 * L, head_length=0.03 * L, fc=c2, ec=c2)

        xmin, xmax, ymin, ymax = self._mesh_bounds(pad_rel=0.05)
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.set_title(field)

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        fig.colorbar(sm, ax=ax, orientation="vertical", label=field)

        if show_springs:
            self._add_springs_to_ax(ax, spring_scale=spring_scale)

        plt.show()

    def plot_stress_along_cut(self, cut_position, cut_direction="x", field="sigma_x",
                              ngrid=250, method="linear",
                              restrict_to_mesh=True, show_cut_line=False):
        """
        Plot along a straight cut line with interpolation of element-center values.

        IMPORTANT FIX:
          If restrict_to_mesh=True, points on the cut that lie outside the actual mesh
          (e.g. holes / gaps in multi-patch) are removed by geometric masking.

        cut_direction:
            "x" -> vertical line x = cut_position (plot vs y)
            "y" -> horizontal line y = cut_position (plot vs x)

        field:
            "sigma_x" | "sigma_y" | "tau_xy" | "n_x" | "n_y" | "n_xy"
        """
        if not hasattr(self.m, "stress_elem_avg"):
            self.m.CalculateInnerElementForces_Gauss()

        field_idx = {
            "sigma_x": 0, "sigma_y": 1, "tau_xy": 2,
            "n_x": 0, "n_y": 1, "n_xy": 2
        }
        if field not in field_idx:
            raise ValueError(f"Unknown field '{field}'. Choose from {list(field_idx.keys())}")

        use_n = field.startswith("n_")
        j = field_idx[field]

        # Element centers and values
        n_elem = len(self.m.Meshing.EL)
        centroids = np.zeros((n_elem, 2), dtype=float)
        vals = np.zeros(n_elem, dtype=float)

        for e, el in enumerate(self.m.Meshing.EL):
            coords = self._element_coords(el)
            centroids[e, :] = coords.mean(axis=0)
            vals[e] = float(self.m.n_elem_avg[e, j] if use_n else self.m.stress_elem_avg[e, j])

        # Interpolation bounds (centroid-based)
        xmin, ymin = centroids.min(axis=0)
        xmax, ymax = centroids.max(axis=0)

        # clamp cut_position
        if cut_direction.lower() == "x":
            cut_position = float(np.clip(cut_position, xmin, xmax))
        elif cut_direction.lower() == "y":
            cut_position = float(np.clip(cut_position, ymin, ymax))
        else:
            raise ValueError("cut_direction must be 'x' or 'y'")

        # Interpolation grid
        grid_x, grid_y = np.mgrid[
            xmin:xmax:complex(ngrid),
            ymin:ymax:complex(ngrid)
        ]
        grid_z = griddata(centroids, vals, (grid_x, grid_y), method=method)

        # Extract cut
        if cut_direction.lower() == "x":
            cut_index = int(np.argmin(np.abs(grid_x[:, 0] - cut_position)))
            cut_vals = grid_z[cut_index, :]
            cut_coords = grid_y[cut_index, :]
            xlabel = "y"
            title = f"{field} along x = {cut_position:.6g}"
        else:
            cut_index = int(np.argmin(np.abs(grid_y[0, :] - cut_position)))
            cut_vals = grid_z[:, cut_index]
            cut_coords = grid_x[:, cut_index]
            xlabel = "x"
            title = f"{field} along y = {cut_position:.6g}"

        # Remove NaNs
        mask = np.isfinite(cut_vals)
        if mask.sum() < 5:
            print("WARNING: Almost no values on cut (NaNs). Try method='nearest' or choose another cut_position.")
            return

        xplot = cut_coords[mask]
        yplot = cut_vals[mask]

        # IMPORTANT: remove points outside actual mesh (holes / gaps)
        if restrict_to_mesh:
            geom_mask = self._mask_cut_points_in_mesh(cut_direction, cut_position, xplot)
            xplot = xplot[geom_mask]
            yplot = yplot[geom_mask]

            if len(xplot) < 5:
                print("WARNING: Cut is mostly outside mesh (after geometric masking). Choose another cut_position.")
                return

        vmin = float(np.min(yplot))
        vmax = float(np.max(yplot))

        plt.figure()
        plt.plot(xplot, yplot, label=title)

        # Min / Max lines
        plt.axhline(vmin, linestyle="--", linewidth=1, label=f"min = {vmin:.4g}")
        plt.axhline(vmax, linestyle="--", linewidth=1, label=f"max = {vmax:.4g}")

        # Info box
        info = f"min = {vmin:.4g}\nmax = {vmax:.4g}"
        plt.gca().text(
            0.02, 0.98, info,
            transform=plt.gca().transAxes,
            va="top",
            ha="left",
            bbox=dict(boxstyle="round", alpha=0.8)
        )

        plt.xlabel(xlabel)
        plt.ylabel(field)
        plt.grid(True)
        plt.legend()
        plt.title(title)
        plt.tight_layout()
        plt.show()

        # Optional: show cut line on mesh (quick visual check)
        if show_cut_line:
            fig, ax = plt.subplots()
            ax.set_aspect("equal", adjustable="box")
            ax.grid(True)

            # mesh outline
            NL = np.asarray(self.m.Meshing.NL, float)
            EL = np.asarray(self.m.Meshing.EL, int)
            for el in EL:
                coords = np.array([NL[nid - 1] for nid in el], float)
                ax.add_patch(patches.Polygon(coords, closed=True, fill=False, edgecolor="0.7", linewidth=1.0))

            xmin2, xmax2, ymin2, ymax2 = self._mesh_bounds(pad_rel=0.05)
            ax.set_xlim(xmin2, xmax2)
            ax.set_ylim(ymin2, ymax2)

            if cut_direction.lower() == "x":
                ax.axvline(cut_position, linestyle="--")
            else:
                ax.axhline(cut_position, linestyle="--")

            plt.show()

    # ---------- springs overlay ----------
    def _draw_springs(self, ax, NL, Lref, spring_scale=0.05, show_k=False):
        """
        Draw spring symbols for boundary conditions from self.A.BC.
        Expects BC columns: No, DOF, cf in [MN/m]  (optional)
        """
        if not hasattr(self.A, "BC"):
            return []

        bc = self.A.BC
        arts = []

        def spring_poly(L=1.0, nzig=6, amp=0.15):
            xs = np.linspace(0, L, 2 * nzig + 1)
            ys = np.zeros_like(xs)
            for k in range(1, len(xs) - 1):
                ys[k] = amp * (1 if k % 2 else -1)
            return np.column_stack([xs, ys])

        for i in range(len(bc)):
            node = bc["No"].iloc[i]
            dof = str(bc["DOF"].iloc[i]).strip().lower()

            if not str(node).isdigit():
                continue
            node = int(node)

            x, y = NL[node - 1]

            Ls = spring_scale * 0.25 * Lref
            amp = 0.10 * Ls
            pts = spring_poly(L=Ls, nzig=5, amp=amp)

            if dof == "x":
                pts[:, 0] *= -1.0
                pts[:, 0] += x
                pts[:, 1] += y
                line, = ax.plot(pts[:, 0], pts[:, 1], linewidth=1.5)
                arts.append(line)

                wall, = ax.plot([x - Ls, x - Ls], [y - 0.15 * Ls, y + 0.15 * Ls], linewidth=2.0)
                arts.append(wall)

                if show_k and "cf in [MN/m]" in bc.columns:
                    k = bc["cf in [MN/m]"].iloc[i]
                    txt = ax.text(x - 1.1 * Ls, y + 0.18 * Ls, f"k={k:g}", fontsize=8, ha="right")
                    arts.append(txt)

            elif dof == "z":
                X = pts[:, 0]
                Y = pts[:, 1]
                pts2 = np.column_stack([-Y, -X])  # rotation -90deg
                pts2[:, 0] += x
                pts2[:, 1] += y
                line, = ax.plot(pts2[:, 0], pts2[:, 1], linewidth=1.5)
                arts.append(line)

                wall, = ax.plot([x - 0.15 * Ls, x + 0.15 * Ls], [y - Ls, y - Ls], linewidth=2.0)
                arts.append(wall)

                if show_k and "cf in [MN/m]" in bc.columns:
                    k = bc["cf in [MN/m]"].iloc[i]
                    txt = ax.text(x + 0.18 * Ls, y - 1.1 * Ls, f"k={k:g}", fontsize=8, va="top")
                    arts.append(txt)

        return arts

    def _add_springs_to_ax(self, ax, spring_scale=1.0):
        NL = np.asarray(self.m.Meshing.NL, dtype=float)
        xmin, xmax, ymin, ymax = self._mesh_bounds(pad_rel=0.0)
        Lref = 0.15 * max(xmax - xmin, ymax - ymin, 1e-12)
        self._draw_springs(ax, NL, Lref, spring_scale=spring_scale)

    def plot_mesh_with_node_ids(self,
                            show_node_ids=True,
                            show_elem_ids=False,
                            show_springs=True,
                            spring_scale=1.0):
        """
        Backward compatibility wrapper.
        """
        return self.plot_mesh(
            show_node_ids=show_node_ids,
            show_elem_ids=show_elem_ids,
            show_springs=show_springs,
            spring_scale=spring_scale
        )


    def plot_load_vector_interactive(self,
                                    scale0=1.0,
                                    scale_max=20.0,
                                    show_node_ids=False,
                                    show_springs=True,
                                    spring_scale=1.0,
                                    show_loaded_labels=True):
        """
        Backward-compatible wrapper for interactive nodal load plotting.
        """
        import numpy as np
        import matplotlib.pyplot as plt
        import matplotlib.patches as patches
        from matplotlib.widgets import Slider

        if not hasattr(self.A, "Load_Vector"):
            raise RuntimeError("Load_Vector not found. Call GenerateLoadVector() first.")

        NL = np.asarray(self.m.Meshing.NL, dtype=float)
        EL = np.asarray(self.m.Meshing.EL, dtype=int)

        Fx = self.A.Load_Vector[::2].astype(float)
        Fy = self.A.Load_Vector[1::2].astype(float)

        nN = NL.shape[0]
        if len(Fx) != nN:
            raise RuntimeError("Load_Vector size does not match number of nodes.")

        Fmag = np.sqrt(Fx**2 + Fy**2)
        Fmax = float(np.max(Fmag)) if np.max(Fmag) > 0 else 1.0

        xmin, ymin = NL.min(axis=0)
        xmax, ymax = NL.max(axis=0)
        Lref = 0.15 * max(xmax - xmin, ymax - ymin, 1e-12)

        fig, ax = plt.subplots()
        plt.subplots_adjust(bottom=0.18)
        ax.set_aspect("equal", adjustable="box")
        ax.grid(True)
        ax.set_title("Nodal load vector")

        ax.set_xlim(xmin - 0.1*Lref, xmax + 0.1*Lref)
        ax.set_ylim(ymin - 0.1*Lref, ymax + 0.1*Lref)

        # mesh
        for el in EL:
            coords = [NL[nid - 1] for nid in el]
            ax.add_patch(patches.Polygon(coords, closed=True, fill=False,
                                        edgecolor="0.7", linewidth=1.0))

        if show_node_ids:
            for i, (x, y) in enumerate(NL, start=1):
                ax.text(x, y, str(i), fontsize=8, ha="right", va="bottom")

        loaded = np.where((np.abs(Fx) > 1e-14) | (np.abs(Fy) > 1e-14))[0]

        arrow_art = [None] * nN
        for i in loaded:
            x, y = NL[i]
            arrow_art[i] = ax.arrow(x, y, 0.0, 0.0,
                                    head_width=0.05*Lref,
                                    head_length=0.07*Lref,
                                    length_includes_head=True)

            if show_loaded_labels:
                ax.text(x, y, f"{i+1}\n({Fx[i]:.2g},{Fy[i]:.2g})",
                        fontsize=7, ha="left", va="bottom")

        if show_springs:
            self._add_springs_to_ax(ax, spring_scale=spring_scale)

        ax_slider = plt.axes([0.15, 0.06, 0.70, 0.03])
        s_scale = Slider(ax_slider, "Scale", 0.0, scale_max, valinit=scale0)

        def _update(scale):
            for i in loaded:
                try:
                    arrow_art[i].remove()
                except Exception:
                    pass

                x, y = NL[i]
                dx = scale * Lref * Fx[i] / Fmax
                dy = scale * Lref * Fy[i] / Fmax

                arrow_art[i] = ax.arrow(
                    x, y, dx, dy,
                    head_width=0.05*Lref,
                    head_length=0.07*Lref,
                    length_includes_head=True
                )

            fig.canvas.draw_idle()

        s_scale.on_changed(_update)
        _update(scale0)

        plt.show()



    def plot_principal_membrane_forces(self,
                                    which="n1",
                                    mode="elem",
                                    draw_dirs=True,
                                    dir_scale=1.0,
                                    show_springs=True,
                                    spring_scale=1.0):
        """
        Plot principal membrane forces.
        which: "n1" or "n2"
        mode : "elem" (uses self.m.n_princ_elem_avg) or "node" (uses self.m.n_princ_node + nodal interpolation)
        draw_dirs: draw principal directions (theta)
        dir_scale: scale factor for direction arrows (relative)
        """

        import numpy as np
        import matplotlib.pyplot as plt
        import matplotlib.patches as patches
        import matplotlib.cm as cm
        import matplotlib.colors as colors

        # ensure principal results exist
        if not hasattr(self.m, "n_princ_elem_avg"):
            # call solver with principal enabled
            self.m.CalculateInnerElementForces_Gauss(compute_nodal=(mode == "node"), compute_principal=True)

        NL = np.asarray(self.m.Meshing.NL, float)
        EL = np.asarray(self.m.Meshing.EL, int)

        idx = 0 if which == "n1" else 1
        title = f"principal membrane force {which}"

        # ----------------------------
        # values per element (preferred)
        # ----------------------------
        if mode.lower() == "elem":
            data = np.asarray(self.m.n_princ_elem_avg, float)  # (n_elem,3) [n1,n2,theta]
            vals = data[:, idx]
            thetas = data[:, 2]

            vmin, vmax = float(np.min(vals)), float(np.max(vals))
            if np.isclose(vmin, vmax):
                vmin -= 1.0
                vmax += 1.0

            norm = colors.Normalize(vmin=vmin, vmax=vmax)
            cmap = cm.viridis

            fig, ax = plt.subplots()
            ax.set_aspect("equal", adjustable="box")
            ax.grid(True)

            # reference length for arrows
            xmin, xmax, ymin, ymax = self._mesh_bounds(pad_rel=0.0)
            Lref = 0.12 * max(xmax - xmin, ymax - ymin, 1e-12) * dir_scale

            for e, el in enumerate(EL):
                coords = np.array([NL[nid - 1] for nid in el], float)
                poly = patches.Polygon(coords, closed=True, edgecolor="k", facecolor=cmap(norm(vals[e])))
                ax.add_patch(poly)

                if draw_dirs:
                    c = coords.mean(axis=0)
                    th = float(thetas[e])

                    # principal direction (theta) and orthogonal direction
                    dx1, dy1 = Lref*np.cos(th), Lref*np.sin(th)
                    dx2, dy2 = -Lref*np.sin(th), Lref*np.cos(th)

                    ax.arrow(c[0], c[1], dx1, dy1, head_width=0.12*Lref, head_length=0.12*Lref,
                            length_includes_head=True)
                    ax.arrow(c[0], c[1], dx2, dy2, head_width=0.12*Lref, head_length=0.12*Lref,
                            length_includes_head=True)

            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
            ax.set_title(title)

            sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
            sm.set_array([])
            fig.colorbar(sm, ax=ax, orientation="vertical", label=which)

            if show_springs:
                self._add_springs_to_ax(ax, spring_scale=spring_scale)

            plt.show()
            return

        # ----------------------------
        # nodal mode (smooth): node values -> element facecolor via averaging
        # ----------------------------
        if mode.lower() == "node":
            if not hasattr(self.m, "n_princ_node"):
                self.m.CalculateInnerElementForces_Gauss(compute_nodal=True, compute_principal=True)

            nnode = np.asarray(self.m.n_princ_node, float)  # (n_nodes,3)
            node_vals = nnode[:, idx]

            # element value = mean of its node values (simple, stable)
            vals = np.zeros((EL.shape[0],), float)
            thetas = np.zeros((EL.shape[0],), float)
            for e, el in enumerate(EL):
                ids = np.array(el, int) - 1
                vals[e] = float(np.mean(node_vals[ids]))
                thetas[e] = float(np.mean(nnode[ids, 2]))  # averaged theta (ok for structured meshes)

            vmin, vmax = float(np.min(vals)), float(np.max(vals))
            if np.isclose(vmin, vmax):
                vmin -= 1.0
                vmax += 1.0

            norm = colors.Normalize(vmin=vmin, vmax=vmax)
            cmap = cm.viridis

            fig, ax = plt.subplots()
            ax.set_aspect("equal", adjustable="box")
            ax.grid(True)

            xmin, xmax, ymin, ymax = self._mesh_bounds(pad_rel=0.0)
            Lref = 0.12 * max(xmax - xmin, ymax - ymin, 1e-12) * dir_scale

            for e, el in enumerate(EL):
                coords = np.array([NL[nid - 1] for nid in el], float)
                poly = patches.Polygon(coords, closed=True, edgecolor="k", facecolor=cmap(norm(vals[e])))
                ax.add_patch(poly)

                if draw_dirs:
                    c = coords.mean(axis=0)
                    th = float(thetas[e])
                    dx1, dy1 = Lref*np.cos(th), Lref*np.sin(th)
                    dx2, dy2 = -Lref*np.sin(th), Lref*np.cos(th)
                    ax.arrow(c[0], c[1], dx1, dy1, head_width=0.12*Lref, head_length=0.12*Lref,
                            length_includes_head=True)
                    ax.arrow(c[0], c[1], dx2, dy2, head_width=0.12*Lref, head_length=0.12*Lref,
                            length_includes_head=True)

            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
            ax.set_title(title)

            sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
            sm.set_array([])
            fig.colorbar(sm, ax=ax, orientation="vertical", label=which)

            if show_springs:
                self._add_springs_to_ax(ax, spring_scale=spring_scale)

            plt.show()
            return

        raise ValueError("mode must be 'elem' or 'node'")



from KIB_LAP.Querschnittswerte import CrossSectionThin
import pandas as pd

# Lade die hochgeladenen CSV-Dateien
node_file_path = 'Querschnittswerte/Knoten.csv'
element_file_path = 'Querschnittswerte/Elemente.csv'

# Dateien einlesen
node_data = pd.read_csv(node_file_path)
element_data = pd.read_csv(element_file_path)

CS = CrossSectionThin(210000,0.30,node_param=node_data, element_param=element_data, Speichername = "Calculation_2")

CS.Calculation_Start()



# ============================================================
# FULL, RUNNABLE SCRIPT (UPDATED)
#   - Loads + reactions + spring reactions + member forces
#   - Sliders + CheckButtons
#   - Optional node numbers + element numbers
#   - Springs plotted ONLY where k>0 (or spring_dofs provided)
# ============================================================

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, CheckButtons
from matplotlib.patches import FancyArrowPatch
from matplotlib.lines import Line2D
from matplotlib import transforms


class PlottingStructure:
    def __init__(
        self,
        _UG_FINAL,
        _FI_FINAL,
        _EXTFORCES,
        _MBRFORCES,
        _nodes,
        _members,
        _P0,
        _restrainedIndex,
        _reactionsFlag=True,
        _forceVector=None,
        _memberType=None,
    ):
        self.UG_FINAL = np.asarray(_UG_FINAL, dtype=float)          # (2*nNodes, nSteps)
        self.FI_FINAL = np.asarray(_FI_FINAL, dtype=float)          # (2*nNodes, nSteps)
        self.EXTFORCES = np.asarray(_EXTFORCES, dtype=float)        # (2*nNodes, nSteps)
        self.MBRFORCES = np.asarray(_MBRFORCES, dtype=float)        # (nMembers, nSteps)

        self.nodes = np.asarray(_nodes, dtype=float)                # (nNodes, 2)
        self.members = list(_members)                               # [(i,j), ...] with 1-based node indices
        self.P0 = np.asarray(_P0, dtype=float)                      # (nMembers,)
        self.restrainedIndex = set(int(i) for i in _restrainedIndex)  # dof indices (0-based)
        self.reactionsFlag = bool(_reactionsFlag)

        self.forceVector = _forceVector
        self.memberType = _memberType if _memberType is not None else ["b"] * len(self.members)

        # Optional (set from outside):
        # self.Kspring_diag : (nDoF,) diagonal spring stiffness in global DOFs
        # self.spring_dofs  : set[int] DOFs that truly have springs (optional)
        self.Kspring_diag = None
        self.spring_dofs = None

        # Tunable parameters
        self.params = {
            "Axial_Forces": False,
            "Show_Loads": True,
            "Show_Reactions": True,
            "Show_Springs": True,          # NEW
            "Show_MemberVectors": True,
            "Show_Legend": True,

            "Show_NodeNumbers": True,      # NEW
            "Show_ElementNumbers": False,  # NEW

            "label_offset": 0.02,          # node number offset (data coords)
            "elem_label_offset": 0.00,     # extra offset for element labels (data coords)
            "xMargin": 0.2,
            "yMargin": 0.4,

            "scaleFactor": 1.0,            # deformation scale
            "Load_Increment": self.UG_FINAL.shape[1] - 1,
            "Final_config": True,

            # label & arrow tuning
            "textScale": 1.0,
            "arrowLenMin": 0.03,           # relative to span
            "arrowLenMax": 0.12,           # relative to span
            "arrowHeadScale": 12.0,        # points
            "labelOffsetPts": 10,          # points
            "refQuantile": 0.95,           # reference force quantile for scaling
            "softPower": 0.65,             # 0.5..0.9 (lower => more compression)
            "collisionIters": 25,
            "collisionRadiusPts": 14,
            "vectorScaleFactor": 1.0,      # member vector scale

            # NEW: spring filter thresholds
            "springK_tol": 1e-12,
            "springF_tol": 1e-9,
        }

    # -----------------------
    # Helpers
    # -----------------------
    @staticmethod
    def _soft_scale(value, ref, Lmin, Lmax, p=0.65):
        if ref <= 0:
            return 0.0
        x = abs(value) / ref
        s = x**p
        s = min(max(s, 0.0), 1.0)
        return Lmin + (Lmax - Lmin) * s

    @staticmethod
    def _auto_fontsize(ax, base=10, min_fs=7, max_fs=13, textScale=1.0):
        bbox = ax.get_window_extent().transformed(ax.figure.dpi_scale_trans.inverted())
        width_inch = bbox.width
        fs = base * (width_inch / 6.0) ** 0.4
        fs *= textScale
        return float(np.clip(fs, min_fs, max_fs))

    def _add_force_arrow(
        self,
        ax,
        x, y,
        vx, vy,
        magnitude,
        color="blue",
        label=None,
        Lmin=0.1, Lmax=0.5,
        ref=1.0,
        lw=1.2,
        fontsize=10,
        text_offset_pts=10,
        head_scale=12.0,
        zorder=5,
    ):
        dir_vec = np.array([vx, vy], dtype=float)
        nrm = np.linalg.norm(dir_vec)
        if nrm < 1e-12:
            return None

        dir_vec /= nrm
        L = self._soft_scale(magnitude, ref, Lmin, Lmax, p=self.params["softPower"])
        dx, dy = dir_vec[0] * L, dir_vec[1] * L

        arrow = FancyArrowPatch(
            (x, y), (x + dx, y + dy),
            arrowstyle="-|>",
            mutation_scale=head_scale,
            linewidth=lw,
            color=color,
            zorder=zorder
        )
        ax.add_patch(arrow)

        txt_artist = None
        if label:
            perp = np.array([-dir_vec[1], dir_vec[0]])
            perp_sign = 1.0
            if (dir_vec[0] < 0) or (dir_vec[1] < 0):
                perp_sign = -1.0

            offset = perp_sign * perp * text_offset_pts
            trans = ax.transData + transforms.ScaledTranslation(
                offset[0] / 72.0, offset[1] / 72.0, ax.figure.dpi_scale_trans
            )

            txt_artist = ax.text(
                x + dx, y + dy, label,
                transform=trans,
                fontsize=fontsize,
                weight="bold",
                va="center", ha="center",
                color=color,
                bbox=dict(facecolor="white", edgecolor="none", alpha=0.80, pad=1.5),
                zorder=zorder + 1
            )

        return txt_artist

    def _resolve_text_collisions(self, ax, text_artists, iters=25, min_dist_pts=14):
        text_artists = [t for t in text_artists if t is not None]
        if not text_artists:
            return

        fig = ax.figure
        fig.canvas.draw()
        renderer = fig.canvas.get_renderer()

        min_dist_px = (min_dist_pts / 72.0) * fig.dpi

        for _ in range(int(iters)):
            moved = False

            centers = []
            for t in text_artists:
                bb = t.get_window_extent(renderer=renderer).expanded(1.05, 1.10)
                centers.append(np.array([bb.x0 + bb.width / 2, bb.y0 + bb.height / 2]))

            for i in range(len(text_artists)):
                for j in range(i + 1, len(text_artists)):
                    d = centers[j] - centers[i]
                    dist = float(np.linalg.norm(d))
                    if dist < min_dist_px:
                        if dist < 1e-6:
                            d = np.array([1.0, 0.0])
                            dist = 1.0
                        push = (min_dist_px - dist) * 0.5
                        step = (d / dist) * push

                        for idx, sgn in [(i, -1.0), (j, +1.0)]:
                            t = text_artists[idx]
                            x, y = t.get_position()
                            p_disp = ax.transData.transform((x, y))
                            p_disp = p_disp + sgn * step
                            p_data = ax.transData.inverted().transform(p_disp)
                            t.set_position((p_data[0], p_data[1]))

                        moved = True

            if not moved:
                break

        fig.canvas.draw_idle()

    def _apply_legend(self, ax):
        proxies = [
            Line2D([0], [0], color="blue", lw=2, label="Loads"),
            Line2D([0], [0], color="black", lw=2, label="Reactions (BC)"),
            Line2D([0], [0], color="purple", lw=2, label="Springs"),
            Line2D([0], [0], color="purple", lw=2, linestyle=":", label="Member vectors"),
            Line2D([0], [0], color="#33cc99", lw=2, linestyle="--", label="Undeformed"),
            Line2D([0], [0], color="#1f77b4", lw=3, label="Deformed (tension)"),
            Line2D([0], [0], color="#d62728", lw=3, label="Deformed (compression)"),
        ]
        ax.legend(handles=proxies, loc="upper right", framealpha=0.9)

    # -----------------------
    # Main plot on axis
    # -----------------------
    def _plot_on_axis(self, ax, **kwargs):
        # params
        Axial_Forces = kwargs["Axial_Forces"]
        Show_Loads = kwargs["Show_Loads"]
        Show_Reactions = kwargs["Show_Reactions"]
        Show_Springs = kwargs["Show_Springs"]
        Show_MemberVectors = kwargs["Show_MemberVectors"]
        Show_Legend = kwargs["Show_Legend"]
        Show_NodeNumbers = kwargs["Show_NodeNumbers"]
        Show_ElementNumbers = kwargs["Show_ElementNumbers"]

        label_offset = kwargs["label_offset"]
        elem_label_offset = kwargs["elem_label_offset"]
        xMargin = kwargs["xMargin"]
        yMargin = kwargs["yMargin"]
        scaleFactor = kwargs["scaleFactor"]
        Load_Increment = int(kwargs["Load_Increment"])
        Final_config = kwargs["Final_config"]

        textScale = kwargs["textScale"]
        arrowLenMin = kwargs["arrowLenMin"]
        arrowLenMax = kwargs["arrowLenMax"]
        arrowHeadScale = kwargs["arrowHeadScale"]
        labelOffsetPts = kwargs["labelOffsetPts"]
        refQuantile = kwargs["refQuantile"]
        collisionIters = kwargs["collisionIters"]
        collisionRadiusPts = kwargs["collisionRadiusPts"]
        vectorScaleFactor = kwargs["vectorScaleFactor"]

        springK_tol = kwargs["springK_tol"]
        springF_tol = kwargs["springF_tol"]

        if Final_config:
            Load_Increment = -1

        ug = np.asarray(self.UG_FINAL[:, Load_Increment]).flatten()
        fi = np.asarray(self.FI_FINAL[:, Load_Increment]).flatten()
        forceVector = np.asarray(self.EXTFORCES[:, Load_Increment]).flatten()
        mbrForces = np.asarray(self.MBRFORCES[:, Load_Increment]).flatten()

        ax.clear()
        ax.set_aspect("equal", adjustable="datalim")
        ax.grid(True)

        # span for relative arrow lengths
        min_x, max_x = self.nodes[:, 0].min(), self.nodes[:, 0].max()
        min_y, max_y = self.nodes[:, 1].min(), self.nodes[:, 1].max()
        span = max(max_x - min_x, max_y - min_y)
        Lmin = max(1e-9, arrowLenMin * span)
        Lmax = max(Lmin * 1.05, arrowLenMax * span)

        # reference force
        allF = np.concatenate([np.abs(forceVector), np.abs(fi), np.abs(mbrForces)])
        allF = allF[allF > 1e-9]
        ref_force = float(np.quantile(allF, refQuantile)) if allF.size else 1.0

        # fontsize
        fs = self._auto_fontsize(ax, base=10, textScale=textScale)

        # --- SPRINGS computed ONCE ---
        u_vec = ug  # same as UG_FINAL[:, step]
        if hasattr(self, "Kspring_diag") and self.Kspring_diag is not None:
            k_diag = np.asarray(self.Kspring_diag, dtype=float).flatten()
            f_spring = -k_diag * u_vec
        else:
            k_diag = None
            f_spring = np.zeros_like(u_vec)

        # DOFs which truly have springs
        if self.spring_dofs is not None:
            spring_dofs = set(int(d) for d in self.spring_dofs)
        elif k_diag is not None:
            spring_dofs = set(np.where(k_diag > springK_tol)[0].tolist())
        else:
            spring_dofs = set()

        # nodes + node labels
        node_texts = []
        for i, (x, y) in enumerate(self.nodes):
            ax.plot(x, y, "o", color="#33cc99", zorder=3)

            if Show_NodeNumbers:
                t = ax.text(
                    x + label_offset, y + label_offset, str(i + 1),
                    fontsize=fs, weight="bold",
                    bbox=dict(facecolor="white", edgecolor="none", alpha=0.6, pad=1.0),
                    zorder=10
                )
                node_texts.append(t)

        # undeformed members
        print(self.members)
        for (ni, nj) in self.members:
            ix, iy = self.nodes[ni - 1]
            jx, jy = self.nodes[nj - 1]
            ax.plot([ix, jx], [iy, jy], "--", color="#33cc99", linewidth=1.5, zorder=1)

        # deformed members (thickness by abs force, color by sign)
        max_force_abs = float(np.max(np.abs(mbrForces))) if np.max(np.abs(mbrForces)) != 0 else 1.0
        member_texts = []

        for e, (ni, nj) in enumerate(self.members):
            ix, iy = self.nodes[ni - 1]
            jx, jy = self.nodes[nj - 1]

            ia = 2 * ni - 2
            ib = 2 * ni - 1
            ja = 2 * nj - 2
            jb = 2 * nj - 1

            ixN = ix + ug[ia] * scaleFactor
            iyN = iy + ug[ib] * scaleFactor
            jxN = jx + ug[ja] * scaleFactor
            jyN = jy + ug[jb] * scaleFactor

            force_abs = abs(mbrForces[e])
            line_width = 1.0 + 4.0 * (force_abs / max_force_abs)

            col = "#1f77b4" if mbrForces[e] >= 0 else "#d62728"
            ax.plot([ixN, jxN], [iyN, jyN], "-", linewidth=line_width, color=col, zorder=2)

            # element numbering (NEW)
            if Show_ElementNumbers:
                mx = 0.5 * (ixN + jxN)
                my = 0.5 * (iyN + jyN)
                t_el = ax.text(
                    mx + elem_label_offset, my + elem_label_offset, f"E{e+1}",
                    fontsize=max(7, fs - 1),
                    weight="bold",
                    bbox=dict(facecolor="white", edgecolor="none", alpha=0.65, pad=1.0),
                    zorder=12,
                    color="purple"
                )
                member_texts.append(t_el)

            if Axial_Forces:
                preTen = self.P0[e] / 1000.0
                axialForce = mbrForces[e] / 1000.0 - preTen
                label = f"{axialForce:.2f} kN\n(+{preTen:.2f} PT)"
                mx = 0.5 * (ixN + jxN)
                my = 0.5 * (iyN + jyN)
                t = ax.text(
                    mx, my, label,
                    fontsize=max(7, fs - 1),
                    weight="bold",
                    bbox=dict(facecolor="white", edgecolor="none", alpha=0.75, pad=1.5),
                    zorder=11
                )
                member_texts.append(t)

        # loads + reactions + springs
        force_texts = []
        nNodes = self.nodes.shape[0]

        for n in range(nNodes):
            node_num = n + 1
            i_hor = 2 * node_num - 2
            i_ver = 2 * node_num - 1

            ix, iy = self.nodes[n]
            ixN = ix + ug[i_hor] * scaleFactor
            iyN = iy + ug[i_ver] * scaleFactor

            # loads
            if Show_Loads:
                if abs(forceVector[i_hor]) > 1e-9:
                    t = self._add_force_arrow(
                        ax, ixN, iyN,
                        np.sign(forceVector[i_hor]), 0.0,
                        magnitude=forceVector[i_hor],
                        color="blue",
                        label=f"{forceVector[i_hor]/1000:.2f} kN",
                        Lmin=Lmin, Lmax=Lmax, ref=ref_force,
                        lw=1.2, fontsize=fs,
                        text_offset_pts=labelOffsetPts,
                        head_scale=arrowHeadScale,
                        zorder=6
                    )
                    if t is not None:
                        force_texts.append(t)

                if abs(forceVector[i_ver]) > 1e-9:
                    t = self._add_force_arrow(
                        ax, ixN, iyN,
                        0.0, np.sign(forceVector[i_ver]),
                        magnitude=forceVector[i_ver],
                        color="blue",
                        label=f"{forceVector[i_ver]/1000:.2f} kN",
                        Lmin=Lmin, Lmax=Lmax, ref=ref_force,
                        lw=1.2, fontsize=fs,
                        text_offset_pts=labelOffsetPts,
                        head_scale=arrowHeadScale,
                        zorder=6
                    )
                    if t is not None:
                        force_texts.append(t)

            # reactions (BC)
            if Show_Reactions:
                if (i_hor in self.restrainedIndex) and abs(fi[i_hor]) > 1e-9:
                    t = self._add_force_arrow(
                        ax, ixN, iyN,
                        np.sign(fi[i_hor]), 0.0,
                        magnitude=fi[i_hor],
                        color="black",
                        label=f"{fi[i_hor]/1000:.2f} kN",
                        Lmin=Lmin, Lmax=Lmax, ref=ref_force,
                        lw=1.4, fontsize=fs,
                        text_offset_pts=labelOffsetPts,
                        head_scale=arrowHeadScale,
                        zorder=7
                    )
                    if t is not None:
                        force_texts.append(t)

                if (i_ver in self.restrainedIndex) and abs(fi[i_ver]) > 1e-9:
                    t = self._add_force_arrow(
                        ax, ixN, iyN,
                        0.0, np.sign(fi[i_ver]),
                        magnitude=fi[i_ver],
                        color="black",
                        label=f"{fi[i_ver]/1000:.2f} kN",
                        Lmin=Lmin, Lmax=Lmax, ref=ref_force,
                        lw=1.4, fontsize=fs,
                        text_offset_pts=labelOffsetPts,
                        head_scale=arrowHeadScale,
                        zorder=7
                    )
                    if t is not None:
                        force_texts.append(t)

            # springs (NEW, independent of reactions)
            if Show_Springs and (k_diag is not None):
                # x spring at this node?
                if (i_hor in spring_dofs) and abs(f_spring[i_hor]) > springF_tol:
                    t = self._add_force_arrow(
                        ax, ixN, iyN,
                        np.sign(f_spring[i_hor]), 0.0,
                        magnitude=f_spring[i_hor],
                        color="purple",
                        label=f"{f_spring[i_hor]/1000:.2f} kN",
                        Lmin=Lmin, Lmax=Lmax, ref=ref_force,
                        lw=1.4, fontsize=fs,
                        text_offset_pts=labelOffsetPts,
                        head_scale=arrowHeadScale,
                        zorder=7
                    )
                    if t is not None:
                        force_texts.append(t)

                # y spring at this node?
                if (i_ver in spring_dofs) and abs(f_spring[i_ver]) > springF_tol:
                    t = self._add_force_arrow(
                        ax, ixN, iyN,
                        0.0, np.sign(f_spring[i_ver]),
                        magnitude=f_spring[i_ver],
                        color="purple",
                        label=f"{f_spring[i_ver]/1000:.2f} kN",
                        Lmin=Lmin, Lmax=Lmax, ref=ref_force,
                        lw=1.4, fontsize=fs,
                        text_offset_pts=labelOffsetPts,
                        head_scale=arrowHeadScale,
                        zorder=7
                    )
                    if t is not None:
                        force_texts.append(t)

        # member vectors (purple, orthogonal)
        if Show_MemberVectors:
            for e, (ni, nj) in enumerate(self.members):
                ix, iy = self.nodes[ni - 1]
                jx, jy = self.nodes[nj - 1]

                ia = 2 * ni - 2
                ib = 2 * ni - 1
                ja = 2 * nj - 2
                jb = 2 * nj - 1

                ixN = ix + ug[ia] * scaleFactor
                iyN = iy + ug[ib] * scaleFactor
                jxN = jx + ug[ja] * scaleFactor
                jyN = jy + ug[jb] * scaleFactor

                direction = np.array([jx - ix, jy - iy], dtype=float)
                L = np.linalg.norm(direction)
                if L < 1e-12:
                    continue

                dir_norm = direction / L
                orth = np.array([-dir_norm[1], dir_norm[0]])

                vec_L = self._soft_scale(mbrForces[e], ref_force, Lmin * 0.7, Lmax * 0.7, p=self.params["softPower"])
                vec_L *= np.sign(mbrForces[e]) * vectorScaleFactor
                vec = orth * vec_L

                for (x0, y0) in [(ixN, iyN), (jxN, jyN)]:
                    a = FancyArrowPatch(
                        (x0, y0), (x0 + vec[0], y0 + vec[1]),
                        arrowstyle="-|>",
                        mutation_scale=max(8.0, 0.8 * arrowHeadScale),
                        linewidth=1.0,
                        color="purple",
                        alpha=0.75,
                        zorder=4
                    )
                    ax.add_patch(a)

                ax.plot(
                    [ixN + vec[0], jxN + vec[0]],
                    [iyN + vec[1], jyN + vec[1]],
                    color="purple",
                    linestyle=":",
                    linewidth=1.0,
                    alpha=0.75,
                    zorder=3
                )

        # axis limits / labels
        ax.set_xlim(min_x - xMargin, max_x + xMargin)
        ax.set_ylim(min_y - yMargin, max_y + yMargin)
        ax.set_xlabel("Distance (m)", fontsize=fs)
        ax.set_ylabel("Distance (m)", fontsize=fs)
        ax.set_title("Deflected shape, forces & reactions", fontsize=fs + 2, weight="bold")

        if Show_Legend:
            self._apply_legend(ax)

        # collision reduction
        self._resolve_text_collisions(
            ax,
            node_texts + member_texts + force_texts,
            iters=collisionIters,
            min_dist_pts=collisionRadiusPts
        )

    # -----------------------
    # Interactive window
    # -----------------------
    def create_structure_plot(self):
        fig = plt.figure(figsize=(11, 9))
        ax_main = fig.add_axes([0.25, 0.35, 0.70, 0.60])
        fig.suptitle("Structure Plot (optimized)", fontsize=14)

        slider_left = 0.25
        slider_width = 0.65
        slider_height = 0.03
        slider_y = 0.28

        def add_slider(label, vmin, vmax, vinit, vstep):
            nonlocal slider_y
            ax_s = fig.add_axes([slider_left, slider_y, slider_width, slider_height])
            s = Slider(ax=ax_s, label=label, valmin=vmin, valmax=vmax, valinit=vinit, valstep=vstep)
            slider_y -= 0.04
            return s

        # Sliders
        s_xMargin = add_slider("xMargin", 0.0, 5.0, self.params["xMargin"], 0.1)
        s_yMargin = add_slider("yMargin", 0.0, 5.0, self.params["yMargin"], 0.1)
        s_scaleFactor = add_slider("scaleFactor", 0.0, 10.0, self.params["scaleFactor"], 0.1)

        s_textScale = add_slider("textScale", 0.6, 1.8, self.params["textScale"], 0.05)
        s_arrowMin = add_slider("arrowLenMin", 0.005, 0.10, self.params["arrowLenMin"], 0.005)
        s_arrowMax = add_slider("arrowLenMax", 0.02, 0.30, self.params["arrowLenMax"], 0.01)
        s_head = add_slider("arrowHeadScale", 6.0, 26.0, self.params["arrowHeadScale"], 1.0)
        s_lblPts = add_slider("labelOffsetPts", 0.0, 25.0, self.params["labelOffsetPts"], 1.0)
        s_vecScale = add_slider("vectorScaleFactor", 0.0, 5.0, self.params["vectorScaleFactor"], 0.1)

        max_load = self.UG_FINAL.shape[1] - 1
        s_load = add_slider("Load_Increment", 0, max_load, self.params["Load_Increment"], 1)

        s_collRad = add_slider("collisionRadiusPts", 6, 30, self.params["collisionRadiusPts"], 1)
        s_collIt = add_slider("collisionIters", 0, 60, self.params["collisionIters"], 1)

        # CheckButtons
        ax_check = fig.add_axes([0.05, 0.38, 0.18, 0.32])
        labels = [
            "Loads", "Reactions", "Springs",
            "MemberVectors", "Axial_Forces",
            "NodeNumbers", "ElementNumbers",
            "Final_config", "Legend"
        ]
        actives = [
            self.params["Show_Loads"],
            self.params["Show_Reactions"],
            self.params["Show_Springs"],
            self.params["Show_MemberVectors"],
            self.params["Axial_Forces"],
            self.params["Show_NodeNumbers"],
            self.params["Show_ElementNumbers"],
            self.params["Final_config"],
            self.params["Show_Legend"]
        ]
        check = CheckButtons(ax_check, labels=labels, actives=actives)

        def redraw():
            self._plot_on_axis(ax_main, **self.params)
            fig.canvas.draw_idle()

        def slider_update(_):
            self.params["xMargin"] = float(s_xMargin.val)
            self.params["yMargin"] = float(s_yMargin.val)
            self.params["scaleFactor"] = float(s_scaleFactor.val)

            self.params["textScale"] = float(s_textScale.val)
            self.params["arrowLenMin"] = float(s_arrowMin.val)
            self.params["arrowLenMax"] = float(s_arrowMax.val)
            self.params["arrowHeadScale"] = float(s_head.val)
            self.params["labelOffsetPts"] = float(s_lblPts.val)
            self.params["vectorScaleFactor"] = float(s_vecScale.val)

            self.params["Load_Increment"] = int(s_load.val)

            self.params["collisionRadiusPts"] = int(s_collRad.val)
            self.params["collisionIters"] = int(s_collIt.val)

            redraw()

        def check_update(label):
            if label == "Loads":
                self.params["Show_Loads"] = not self.params["Show_Loads"]
            elif label == "Reactions":
                self.params["Show_Reactions"] = not self.params["Show_Reactions"]
            elif label == "Springs":
                self.params["Show_Springs"] = not self.params["Show_Springs"]
            elif label == "MemberVectors":
                self.params["Show_MemberVectors"] = not self.params["Show_MemberVectors"]
            elif label == "Axial_Forces":
                self.params["Axial_Forces"] = not self.params["Axial_Forces"]
            elif label == "NodeNumbers":
                self.params["Show_NodeNumbers"] = not self.params["Show_NodeNumbers"]
            elif label == "ElementNumbers":
                self.params["Show_ElementNumbers"] = not self.params["Show_ElementNumbers"]
            elif label == "Final_config":
                self.params["Final_config"] = not self.params["Final_config"]
            elif label == "Legend":
                self.params["Show_Legend"] = not self.params["Show_Legend"]

            redraw()

        # connect events
        for s in [
            s_xMargin, s_yMargin, s_scaleFactor,
            s_textScale, s_arrowMin, s_arrowMax,
            s_head, s_lblPts, s_vecScale,
            s_load, s_collRad, s_collIt
        ]:
            s.on_changed(slider_update)

        check.on_clicked(check_update)

        redraw()
        plt.show()



# DEPENDENCIES
import copy #Allows us to create copies of objects in memory
import math #Math functionality
import numpy as np #Numpy for working with arrays
import matplotlib.pyplot as plt #Plotting functionality 
import matplotlib.colors #For colormap functionality
import ipywidgets as widgets
from glob import glob #Allows check that file exists before import
from numpy import genfromtxt #For importing structure data from csv
import pandas as pd

class Material:
    def __init__(self):
        print("Material Class")
        self.MaterialData()

    def MaterialData(self):
        #MANDATORY IMPORT: nodal coordinates
        if glob('data/Material.csv'): 
            self.Materials = pd.read_csv('data/Material.csv') 
            print('1. 🟢 Material.csv imported')
        else: 
            print('1. 🛑 STOP: Material.csv not found')

        print(self.Materials)
        self.N = self.Materials['No'].to_numpy()
        self.E = self.Materials['E'].to_numpy()
        self.A = self.Materials['A'].to_numpy()
        self.gamma = self.Materials['gamma'].to_numpy()
        self.P0 = self.Materials['P0'].to_numpy()



# DEPENDENCIES
import copy  # Allows us to create copies of objects in memory
import math  # Math functionality
import numpy as np  # Numpy for working with arrays
import matplotlib.pyplot as plt  # Plotting functionality
import matplotlib.colors  # For colormap functionality
import ipywidgets as widgets
from glob import glob  # Allows check that file exists before import
from numpy import genfromtxt  # For importing structure data from csv
import pandas as pd


class Rope_Elements_III:
    def __init__(self, InputData):
        print("Rope elements")
        self.Inp = InputData

    def calculateTransMatrix(self, posI, posJ):
        """
        Takes in the position of node I and J and returns the transformation matrix for the member
        This will to be recalculated as the structure deflects with each iteration
        """
        T = np.zeros([2, 4])
        ix = posI[0]  # x-coord for node i
        iy = posI[1]  # y-coord for node i
        jx = posJ[0]  # x-coord for node j
        jy = posJ[1]  # y-coord for node j

        dx = jx - ix  # x-component of vector along member
        dy = jy - iy  # y-component of vector along member
        length = math.sqrt(dx**2 + dy**2)  # Magnitude of vector (length of member)

        lp = dx / length
        mp = dy / length
        lq = -mp
        mq = lp

        T = np.array([[-lp, -mp, lp, mp], [-lq, -mq, lq, mq]])

        return T

    def buildElementStiffnessMatrix(self, n, UG, TMs, lengths, P0, E, Areas):
        """
        Build element stiffness matrix based on current position and axial force
        n = member index
        UG = vector of global cumulative displacements
        """

        # Calculate 'new' positions of nodes using UG
        node_i = self.Inp.members[n][0]  # Node number for node i of this member
        node_j = self.Inp.members[n][1]  # Node number for node j of this member

        # Index of DoF for this member
        ia = 2 * node_i - 2  # horizontal DoF at node i of this member
        ib = 2 * node_i - 1  # vertical DoF at node i of this member
        ja = 2 * node_j - 2  # horizontal DoF at node j of this member
        jb = 2 * node_j - 1  # vertical DoF at node j of this member

        # Displacements
        d_ix = UG[ia, 0]
        d_iy = UG[ib, 0]
        d_jx = UG[ja, 0]
        d_jy = UG[jb, 0]

        # Extract current version of transformation matrix [T]
        TM = TMs[n, :, :]

        # Calculate local displacements [u, v, w] using global cumulative displacements UG
        localDisp = np.matmul(TM, np.array([[d_ix, d_iy, d_jx, d_jy]]).T)
        u = localDisp[0].item()
        v = localDisp[1].item()

        # Calculate extension, e
        Lo = lengths[n]
        e = math.sqrt((Lo + u) ** 2 + v**2) - Lo

        # Calculate matrix [AA]
        a1 = (Lo + u) / (Lo + e)
        a2 = v / (Lo + e)
        AA = np.array([[a1, a2]])

        # Calculate axial load, P

        P = P0[n] + (E[n] * Areas[n] / Lo) * e

        # Calculate matrix [d]
        d11 = P * v**2
        d12 = -P * v * (Lo + u)
        d21 = -P * v * (Lo + u)
        d22 = P * (Lo + u) ** 2
        denominator = (Lo + e) ** 3

        d = (1 / denominator) * np.array([[d11, d12], [d21, d22]])

        # Calculate element stiffness matrix

        NL = np.matrix((AA.T * (E[n] * Areas[n] / Lo) * AA) + d)
        k = TM.T * NL * TM

        # Return element stiffness matrix in quadrants
        K11 = k[0:2, 0:2]
        K12 = k[0:2, 2:4]
        K21 = k[2:4, 0:2]
        K22 = k[2:4, 2:4]

        return [K11, K12, K21, K22]


import numpy as np
import math


class BarElements_I:
    """
    2D Bar/Truss element (Theorie I. Ordnung, small displacement).
    2 DOF per node: ux, uy

    - Uses UNDEFORMED geometry for transformation and stiffness (linear).
    - Optional initial axial force P0 can be included in member force reporting,
      but is NOT used as geometric stiffness here (Theorie I. Ordnung).
    """

    def __init__(self, InputData):
        self.Inp = InputData

        # Precompute direction cosines + lengths from undeformed geometry (constant)
        self.L0 = np.zeros(len(self.Inp.members), dtype=float)
        self.c  = np.zeros(len(self.Inp.members), dtype=float)
        self.s  = np.zeros(len(self.Inp.members), dtype=float)

        for n, (ni, nj) in enumerate(self.Inp.members):
            ix, iy = self.Inp.nodes[ni - 1, 0], self.Inp.nodes[ni - 1, 1]
            jx, jy = self.Inp.nodes[nj - 1, 0], self.Inp.nodes[nj - 1, 1]
            dx, dy = (jx - ix), (jy - iy)
            L = math.sqrt(dx * dx + dy * dy)
            if L == 0.0:
                raise ValueError(f"Bar element has zero length at member index {n} (nodes {ni}-{nj})")

            self.L0[n] = L
            self.c[n]  = dx / L
            self.s[n]  = dy / L

    def buildElementStiffnessMatrix(self, n, UG, TMs_unused, lengths_unused, P0, E, Areas):
        """
        Returns [K11, K12, K21, K22] (each 2x2) in GLOBAL coordinates.
        """
        L  = self.L0[n]
        EA = E[n] * Areas[n]
        k0 = EA / L

        c = self.c[n]
        s = self.s[n]

        # 4x4 global stiffness for 2D truss
        k = k0 * np.array([
            [ c*c,  c*s, -c*c, -c*s],
            [ c*s,  s*s, -c*s, -s*s],
            [-c*c, -c*s,  c*c,  c*s],
            [-c*s, -s*s,  c*s,  s*s]
        ], dtype=float)

        K11 = k[0:2, 0:2]
        K12 = k[0:2, 2:4]
        K21 = k[2:4, 0:2]
        K22 = k[2:4, 2:4]
        return [K11, K12, K21, K22]

    def axial_force(self, n, UG, E, Areas, P0=None):
        """
        Member axial force (tension +) from SMALL displacement theory:
            N = (EA/L) * ( [-c -s c s] * u_e )
        Optionally + P0[n] if you want to report pretension as part of N.
        """
        ni, nj = self.Inp.members[n]
        ia, ib = 2 * ni - 2, 2 * ni - 1
        ja, jb = 2 * nj - 2, 2 * nj - 1

        ue = np.array([UG[ia, 0], UG[ib, 0], UG[ja, 0], UG[jb, 0]], dtype=float)

        c = self.c[n]
        s = self.s[n]
        L = self.L0[n]
        EA = E[n] * Areas[n]

        N = (EA / L) * (-c * ue[0] - s * ue[1] + c * ue[2] + s * ue[3])

        if P0 is not None:
            N = N + float(P0[n])
        return float(N)

    def internal_nodal_forces_global(self, n, UG, E, Areas, P0=None):
        """
        Equivalent internal nodal force vector (4x1) in global coords:
            f_int_e = N * [-c, -s, c, s]^T
        """
        c = self.c[n]
        s = self.s[n]
        N = self.axial_force(n, UG, E, Areas, P0=P0)

        f = N * np.array([[-c], [-s], [c], [s]], dtype=float)
        return f



class Rope_Elements_II:
    def __init__(self):
        print("Rope elements")





import numpy as np
from math import sin, cos
from matplotlib import pyplot as plt


class TwoMassesSpringRopeRK4:
    def __init__(
        self,
        r_inp=5,
        E_inp=30000e6,
        I_inp=833.33,
        l_inp=30,
        m1_inp=5.625e6,
        m2_inp=981250,
        dtinp = 1e-3, timeinp = 10,
        _phi_0 = 30 , _x1_0 = 0,
        _dphi_0 = 0 , _dx1_0 = 0
    ):
        self.r = r_inp
        E = E_inp
        I = I_inp
        l = l_inp


        self.k_1 = 3 * E * I / l**3  

        # Mass parameters
        self.m_1 = m1_inp
        self.m_2 = m2_inp

        # Damping parameters

        # Gravity

        self.g = 9.81  # Gravity constant

        # Initial conditions
        self.x1_0 = _x1_0
        self.phi_0 =  _phi_0  / 180 * np.pi
        self.xd1_0 =  _dphi_0
        self.phid_0 = _x1_0

        # Time steps
        self.dt = dtinp
        self.T_ges = timeinp
        self.N_steps = int(self.T_ges / self.dt)

    def construct_mass_matrix(self, phi):
        m_11 = self.m_1 + self.m_2
        m_12 = self.m_2 * self.r * cos(phi)
        m_21 = self.m_2 * self.r * cos(phi)
        m_22 = self.m_2 * self.r**2

        M = np.array([[m_11, m_12], [m_21, m_22]])
        return M

    def construct_force_vector(self, phi, phid):
        f_10 = -self.k_1 * self.x1_0 + self.m_2 * self.r * phid**2 * sin(phi)
        f_20 = -self.m_2 * self.g * self.r * sin(phi)

        F = np.array([f_10, f_20])
        return F

    def equations_of_motion(self, state):
        x1, phi, xd1, phid = state
        M = self.construct_mass_matrix(phi)
        F = self.construct_force_vector(phi, phid)
        acc = np.linalg.solve(M, F)
        return np.array([xd1, phid, acc[0], acc[1]])

    def RK4_step(self, state):
        k1 = self.equations_of_motion(state)
        k2 = self.equations_of_motion(state + 0.5 * k1 * self.dt)
        k3 = self.equations_of_motion(state + 0.5 * k2 * self.dt)
        k4 = self.equations_of_motion(state + k3 * self.dt)
        return state + self.dt * (k1 + 2 * k2 + 2 * k3 + k4) / 6

    def solve(self):
        state = np.array(
            [self.x1_0, self.phi_0, self.xd1_0, self.phid_0]
        )  # Initial state [x1, phi, xd1, phid]
        list_x1 = []
        list_phi = []
        time = []

        for i in range(self.N_steps):
            state = self.RK4_step(state)
            list_x1.append(state[0])
            list_phi.append(state[1])
            time.append(self.dt * i)

        return time, list_x1, list_phi





try:
    from Element_Stiffness import *
    from Assemble_Stiffness import *
    from Meshing import *
    from Output import *
    from Plotting import *
    from Shell_Calculation import *
except:
    from KIB_LAP.Scheibe.Element_Stiffness import *
    from KIB_LAP.Scheibe.Assemble_Stiffness import *
    from KIB_LAP.Scheibe.Meshing import *
    from KIB_LAP.Scheibe.Output import *
    from KIB_LAP.Scheibe.Plotting import *
    from KIB_LAP.Scheibe.Shell_Calculation import *


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import pandas as pd


class MeshingClass:
    def __init__(self, d1, d2, num_x, num_y, elem_type):
        self._d1 = d1
        self._d2 = d2
        self._num_x = num_x
        self._num_y = num_y
        self._elem_type = elem_type
        self.PD = 2  # Problem-Dimension
        self.NoN = 0  # Number of Nodes NoN
        self.NoE = 0  # Number of Elements NoE
        self.NPE = 4 if elem_type == "rect" else 3  # Nodes per elements
        self.q = np.zeros((4, 2))
        self.NL = np.zeros((1, self.PD))
        self.EL = np.zeros((1, self.NPE))
        self.a = 0
        self.b = 0
        self.element_edges = {}  # Dictionary to store edges by element

    def get_NL(self):
        return self.NL

    def get_EL(self):
        return self.EL

    def get_NPE(self):
        return self.NPE

    def general_mesh_input(self):
        self.q = np.array([[0, 0], [self._d1, 0], [0, self._d2], [self._d1, self._d2]])
        self.NoN = (self._num_x + 1) * (self._num_y + 1)
        self.NoE = self._num_x * self._num_y
        self.NL = np.zeros((self.NoN, self.PD))
        self.EL = np.zeros((self.NoE, self.NPE), dtype=int)
        self.a = (self.q[1, 0] - self.q[0, 0]) / self._num_x
        self.b = (self.q[2, 1] - self.q[0, 1]) / self._num_y

    def generating_rectangular_mesh(self):
        self.general_mesh_input()
        n = 0
        for i in range(1, self._num_y + 2):
            for j in range(1, self._num_x + 2):
                self.NL[n, 0] = self.q[0, 0] + (j - 1) * self.a
                self.NL[n, 1] = self.q[0, 1] + (i - 1) * self.b
                n += 1

        for i in range(1, self._num_y + 1):
            for j in range(1, self._num_x + 1):
                if j == 1:
                    self.EL[(i - 1) * self._num_x + j - 1, 0] = (i - 1) * (
                        self._num_x + 1
                    ) + j
                    self.EL[(i - 1) * self._num_x + j - 1, 1] = (
                        self.EL[(i - 1) * self._num_x + j - 1, 0] + 1
                    )
                    self.EL[(i - 1) * self._num_x + j - 1, 3] = (
                        self.EL[(i - 1) * self._num_x + j - 1, 0] + self._num_x + 1
                    )
                    self.EL[(i - 1) * self._num_x + j - 1, 2] = (
                        self.EL[(i - 1) * self._num_x + j - 1, 3] + 1
                    )
                else:
                    self.EL[(i - 1) * self._num_x + j - 1, 0] = self.EL[
                        (i - 1) * self._num_x + j - 2, 1
                    ]
                    self.EL[(i - 1) * self._num_x + j - 1, 3] = self.EL[
                        (i - 1) * self._num_x + j - 2, 2
                    ]
                    self.EL[(i - 1) * self._num_x + j - 1, 1] = (
                        self.EL[(i - 1) * self._num_x + j - 1, 0] + 1
                    )
                    self.EL[(i - 1) * self._num_x + j - 1, 2] = (
                        self.EL[(i - 1) * self._num_x + j - 1, 3] + 1
                    )

        self.generate_edges()

    def generating_quadrilateral_mesh(self, q):
        """
        Structured Q4 mesh on a general convex quadrilateral using bilinear mapping.

        q: array-like shape (4,2) with corner points in boundary order.
        Recommended order: [q1,q2,q3,q4] counter-clockwise (CCW).
        Example: q1 bottom-left, q2 bottom-right, q3 top-right, q4 top-left.
        """
        q = np.asarray(q, dtype=float)
        if q.shape != (4, 2):
            raise ValueError("q must have shape (4,2)")

        # Store corners
        self.q = q.copy()

        # Setup sizes
        self.NoN = (self._num_x + 1) * (self._num_y + 1)
        self.NoE = self._num_x * self._num_y
        self.NL = np.zeros((self.NoN, self.PD))
        self.EL = np.zeros((self.NoE, self.NPE), dtype=int)

        # For plotting margins: use average edge lengths / divisions
        Lx = 0.5 * (np.linalg.norm(q[1] - q[0]) + np.linalg.norm(q[2] - q[3]))
        Ly = 0.5 * (np.linalg.norm(q[3] - q[0]) + np.linalg.norm(q[2] - q[1]))
        self.a = Lx / self._num_x
        self.b = Ly / self._num_y

        # Bilinear map from (s,t) in [0,1]^2
        q1, q2, q3, q4 = q[0], q[1], q[2], q[3]

        def map_st(s, t):
            return (1 - s) * (1 - t) * q1 + s * (1 - t) * q2 + s * t * q3 + (1 - s) * t * q4

        # Create nodes (same numbering scheme as your rectangular mesh)
        n = 0
        for i in range(0, self._num_y + 1):
            t = i / self._num_y if self._num_y > 0 else 0.0
            for j in range(0, self._num_x + 1):
                s = j / self._num_x if self._num_x > 0 else 0.0
                xy = map_st(s, t)
                self.NL[n, 0] = xy[0]
                self.NL[n, 1] = xy[1]
                n += 1

        # Create elements connectivity (identical pattern as rectangular case)
        # Node numbering: row-major
        for i in range(1, self._num_y + 1):
            for j in range(1, self._num_x + 1):
                # bottom-left node id in this cell
                n1 = (i - 1) * (self._num_x + 1) + j
                n2 = n1 + 1
                n4 = n1 + (self._num_x + 1)
                n3 = n4 + 1

                e = (i - 1) * self._num_x + (j - 1)
                self.EL[e, :] = [n1, n2, n3, n4]

        self.generate_edges()

    def read_q_from_csv(self, filepath="Meshing_Params_q.csv"):
        """
        Liest 4 Eckpunkte q aus CSV mit Spalten 'x[m]' und 'y[m]'.
        Rückgabe: np.array shape (4,2)
        """
        df = pd.read_csv(filepath)

        # toleranter Umgang mit Spaltennamen (falls mal ohne [m])
        cols = [c.strip() for c in df.columns]
        df.columns = cols

        xcol = "x[m]" if "x[m]" in df.columns else ("x" if "x" in df.columns else None)
        ycol = "y[m]" if "y[m]" in df.columns else ("y" if "y" in df.columns else None)
        if xcol is None or ycol is None:
            raise ValueError(f"CSV muss Spalten 'x[m]'/'y[m]' oder 'x'/'y' enthalten. Gefunden: {list(df.columns)}")

        if len(df) != 4:
            raise ValueError(f"CSV muss genau 4 Punkte enthalten, gefunden: {len(df)}")

        q = df[[xcol, ycol]].to_numpy(dtype=float)
        return q

    def generating_quadrilateral_mesh_from_csv(self, filepath="Meshing_Params_q.csv"):
        """
        Convenience: liest q aus CSV und erzeugt direkt das Mesh.
        """
        q = self.read_q_from_csv(filepath)
        self.generating_quadrilateral_mesh(q)

    def generate_edges(self):
        for el_index, el in enumerate(self.EL):
            edges = {
                "left": (el[0], el[3]),
                "bottom": (el[0], el[1]),
                "right": (el[1], el[2]),
                "top": (el[2], el[3]),
            }
            for orientation in edges:
                edges[orientation] = tuple(sorted(edges[orientation]))
            self.element_edges[el_index] = edges

    def _tol(self):
        # typische Elementgröße als Toleranzmaß
        a = float(getattr(self, "a", 0.0))
        b = float(getattr(self, "b", 0.0))
        h = max(a, b, 1e-12)
        return 1e-6 * h  # z.B. 1e-6 der Elementgröße

    def get_left_border_nodes(self, tol=None):
        NL = np.asarray(self.NL, dtype=float)
        if tol is None:
            tol = self._tol()
        x_min = float(NL[:, 0].min())
        return [i + 1 for i, x in enumerate(NL[:, 0]) if abs(x - x_min) <= tol]

    def get_right_border_nodes(self, tol=None):
        NL = np.asarray(self.NL, dtype=float)
        if tol is None:
            tol = self._tol()
        x_max = float(NL[:, 0].max())
        return [i + 1 for i, x in enumerate(NL[:, 0]) if abs(x - x_max) <= tol]

    def get_bottom_border_nodes(self, tol=None):
        NL = np.asarray(self.NL, dtype=float)
        if tol is None:
            tol = self._tol()
        y_min = float(NL[:, 1].min())
        return [i + 1 for i, y in enumerate(NL[:, 1]) if abs(y - y_min) <= tol]

    def get_top_border_nodes(self, tol=None):
        NL = np.asarray(self.NL, dtype=float)
        if tol is None:
            tol = self._tol()
        y_max = float(NL[:, 1].max())
        return [i + 1 for i, y in enumerate(NL[:, 1]) if abs(y - y_max) <= tol]

    def get_element_edges(self):
        return self.element_edges

    def get_coordinate_x(self, node_nr):
        for i in range(0, len(self.EL), 1):
            for j in range(0, len(self.EL[0]), 1):
                index = int(self.EL[i][j])

                if index == node_nr:
                    return self.NL[index - 1][0]

    def get_coordinate_z(self, node_nr):
        for i in range(0, len(self.EL), 1):
            for j in range(0, len(self.EL[0]), 1):
                index = int(self.EL[i][j])

                if index == node_nr:
                    return self.NL[index - 1][1]


    # ============================================================
    # MERGE / STITCH SUPPORT
    # ============================================================

    def merge_with_mesh(self, other, tol=None):
        """
        Vereint self-Mesh und other-Mesh, indem Knoten mit gleichen Koordinaten
        (bis tol) zusammengeführt werden. EL bleibt 1-based.
        """
        if tol is None:
            tol = max(self._tol(), other._tol())

        NL_A = np.asarray(self.NL, float)
        EL_A = np.asarray(self.EL, int)
        NL_B = np.asarray(other.NL, float)
        EL_B = np.asarray(other.EL, int)

        # spatial hash on A
        def key(pt):
            return (round(pt[0] / tol), round(pt[1] / tol))

        hashA = {}
        for i, pt in enumerate(NL_A, start=1):  # 1-based ids
            hashA[key(pt)] = i

        NL = NL_A.tolist()
        mapB = {}  # old B node id -> merged node id

        for j, pt in enumerate(NL_B, start=1):
            k = key(pt)
            if k in hashA:
                mapB[j] = hashA[k]
            else:
                NL.append([pt[0], pt[1]])
                mapB[j] = len(NL)

        # remap B elements
        EL_B_new = np.zeros_like(EL_B)
        for e in range(EL_B.shape[0]):
            for a in range(EL_B.shape[1]):
                EL_B_new[e, a] = mapB[int(EL_B[e, a])]

        # combine
        self.NL = np.asarray(NL, float)
        self.EL = np.vstack([EL_A, EL_B_new]).astype(int)

        self.NoN = len(self.NL)
        self.NoE = len(self.EL)

        # edges aktualisieren
        self.element_edges = {}
        self.generate_edges()

        # optional: mapping speichern
        self.merge_map_other_nodes = mapB


    # in MeshingClass ergänzen
    def translate(self, dx=0.0, dy=0.0):
        self.NL[:, 0] += float(dx)
        self.NL[:, 1] += float(dy)


    def plot_mesh(self,
                show_node_numbers=True,
                show_element_numbers=False,
                node_fontsize=7,
                elem_fontsize=8,
                ax=None):
        """
        Plottet das aktuelle Mesh (NL, EL) inkl. optionaler
        Knoten- und Elementnummern.
        """

        NL = np.asarray(self.NL, float)
        EL = np.asarray(self.EL, int)

        if ax is None:
            fig, ax = plt.subplots(figsize=(7, 6))

        # -------------------------
        # Elemente zeichnen
        # -------------------------
        for e, el in enumerate(EL):
            node_ids = el - 1  # 0-based
            coords = NL[node_ids, :]

            poly = np.vstack([coords, coords[0]])
            ax.plot(poly[:, 0], poly[:, 1], color="black", linewidth=0.8)

            if show_element_numbers:
                xc = coords[:, 0].mean()
                yc = coords[:, 1].mean()
                ax.text(
                    xc, yc, f"E{e+1}",
                    color="red",
                    fontsize=elem_fontsize,
                    ha="center", va="center",
                    bbox=dict(facecolor="white", edgecolor="none", alpha=0.6)
                )

        # -------------------------
        # Knoten plotten
        # -------------------------
        ax.scatter(NL[:, 0], NL[:, 1], s=10, color="black", zorder=3)

        if show_node_numbers:
            for i, (x, y) in enumerate(NL):
                ax.text(
                    x, y, f"N{i+1}",
                    color="blue",
                    fontsize=node_fontsize,
                    ha="right", va="bottom",
                    zorder=4,
                    bbox=dict(facecolor="white", edgecolor="none", alpha=0.6)
                )

        # -------------------------
        # Plot-Layout
        # -------------------------
        ax.set_aspect("equal", adjustable="box")
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.grid(True, linestyle="--", alpha=0.4)
        ax.set_title("Finite-Element-Netz")

        plt.tight_layout()
        plt.show()



import numpy as np

class FFTAnalyzer:
    def __init__(self, t: np.ndarray, y: np.ndarray, window: str = None):
        """
        Parameters
        ----------
        t : np.ndarray
            Time array (must be uniformly spaced).
        y : np.ndarray
            Signal array, same length as t.
        window : str or None
            If not None, must be one of np’s window functions, e.g. 'hann','hamming','blackman'.
        """

        # sampling
        dt = t[1] - t[0]
        if not np.allclose(np.diff(t), dt, rtol=1e-5, atol=1e-8):
            raise ValueError("t must be uniformly spaced")
        self.fs = 1.0 / dt
        self.N  = len(t)
        self._t = t
        self._y = y
        # window
        if window is not None:
            try:
                win = getattr(np, window)(self.N)
            except AttributeError:
                raise ValueError(f"Unknown window '{window}'")
        else:
            win = np.ones(self.N)
        self._y_win = y * win

        # placeholders
        self.freqs    = None
        self.amp      = None
        self.phase    = None
        self._computed = False

    def compute(self):
        """Compute the one-sided FFT, filling freqs, amp, phase."""
        # full FFT
        Y = np.fft.rfft(self._y_win)
        # frequencies
        self.freqs = np.fft.rfftfreq(self.N, d=1/self.fs)
        # amplitude correction (accounting for window & symmetry)
        # multiply by 2/N (except DC and Nyquist if present)
        A = np.abs(Y) * 2.0 / self.N
        A[0] /= 2.0
        if self.N % 2 == 0:  # Nyquist freq at end
            A[-1] /= 2.0
        self.amp   = A
        # phase
        self.phase = np.angle(Y)
        self._computed = True
        return self.freqs, self.amp, self.phase

    def plot(self, ax=None, xlim=None):
        """
        Quick plot of amplitude spectrum. 
        Returns (fig, ax).

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to plot into. Wenn None, wird ein neues Figure/Axes‐Paar erzeugt.
        xlim : tuple of float (xmin, xmax), optional
            Frequenzbereich in Hz für die x‐Achse.
        """
        import matplotlib.pyplot as plt
        if not self._computed:
            self.compute()
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        ax.plot(self.freqs, self.amp)
        ax.set_xlabel("Frequency [Hz]")
        ax.set_ylabel("Amplitude")
        ax.grid(True)

        if xlim is not None:
            ax.set_xlim(xlim)

        return fig, ax




# results.py
from dataclasses import dataclass
import numpy as np
from typing import Any


@dataclass
class AnalysisResults:
    """
    Reiner Datencontainer für Analyse-Ergebnisse.
    KEINE Berechnungen, KEIN Plotting.
    """

    # Input (Geometrie, Materialien, Knoten, Stäbe)
    Inp: Any

    # Globale Ergebnisse
    u_ges: np.ndarray
    GesMat: np.ndarray
    FGes: np.ndarray

    # Elementweise Matrizen
    TransMats: np.ndarray
    K_el_i_store: np.ndarray
    u_el: np.ndarray
    s_el: np.ndarray

    # Schnittgrößen an den Elementenden
    N_el_i_store: np.ndarray
    VY_el_i_store: np.ndarray
    VZ_el_i_store: np.ndarray
    MX_el_i_store: np.ndarray
    MY_el_i_store: np.ndarray
    MZ_el_i_store: np.ndarray

    # Zusatzinfos
    member_length: np.ndarray



from Iteration import IterationClass
from Plotting import PlottingStructure


Iteration = IterationClass()
try:
    Iteration.Summarize()
except:
    Iteration.SummarizeLinear()

plotter = PlottingStructure(
    Iteration.UG_FINAL,
    Iteration.FI_FINAL,
    Iteration.EXTFORCES,
    Iteration.MBRFORCES,
    Iteration.Inp.nodes,
    Iteration.Inp.members,
    Iteration.Mat.P0,
    Iteration.Inp.restrainedIndex,
    _reactionsFlag=True,
)

plotter.Kspring_diag = Iteration.Kspring_diag  # (nDoF,)
# optional (noch besser, damit nur echte Federn geplottet werden)
# plotter.spring_dofs = Iteration.spring_dofs

plotter.create_structure_plot()
